"""
Marching front algorithm to select points from a topo file based on
various criteria.  See select_by_flooding docstring 

"""

from __future__ import print_function
from pylab import *
import sys
from clawpack.geoclaw import topotools



def select_by_flooding(Ztopo, mask=None, prev_pts_chosen=None,
                       Z1=-5., Z2=0., max_iters=None, verbose=False):
    """
    Uses Ztopo as the topography DEM.

    Returns pts_chosen, a 1/0 array of the same shape as Ztopo, indicating
    whether each point is chosen (1) or not (0), based on various criteria.

    If Z1 <= Z2 then points where Ztopo < Z1 are first selected and 
    then a marching algorithm is used to select neighboring points that 
    satisfy Ztopo < Z2. 

    Think of chosen points as "wet" and unchosen points as "dry" 
    and new points are flooded if at least one neighbor is "wet" and 
    the topography is low enough. Starting from deep water (e.g. Z1 = -5)
    this allows flooding up to MHW (Z2 = 0) without going past dikes that
    protect dry land with lower elevation.
     
    However, this can also be called with Z1 > Z2, in which case points
    where Ztopo > Z1 are first selected and then the marching algorithm is 
    used to select neighboring points that satisfy Ztopo > Z2.
    This is useful to select offshore points where the water is shallow,
    (and that are connected to the shore by a path of shallow points) 
    e.g. on the continental shelf. 

    If max_iters=None then iterate to convergence.

    Otherwise, at most max_iters iterations are taken, so setting
    this to a small value only selects points within this many grid
    points of where Ztopo < Z1, useful for buffering.

    If prev_pts_chosen is None we are starting from scratch, otherwise
    we possibly add additional chosen points to an existing array.
    Points where prev_pts_chosen[i,j]==1 won't change but those ==0 may be
    changed to 1 based on the new criteria.

    If mask==None or mask==False then the entire array is subject to 
    selection.  If mask is an array with mask[i,j]==True at some points,
    then either:
        - these masked points are marked as not selected (pts_chosen[i,j]=0)
          if prev_pts_chosen==None, or
        - these masked points are not touched if prev_pts_chosen is an
          array (so previous 0/1 values are preserved).

    Basic strategy (assuming Z1 <= Z2):
        Outside the masked region, if any, initialize pt_chosen to
            pt_chosen=1 where Ztopo < Z1
            pt_chosen=0 where Ztopo >= Z2
            pt_chosen=-5 for intermediate points to indicate unknown        
        Let water intrude to level Z2
            (setting pt_chosen=1 where Ztopo < Z2 and a neighbor is wet)
            This is done by a fast marching iteration in which we keep track 
            of the cells at the front to check next, those for which pt_chosen==1
            and at least one neighbor has pt_chosen=-5 (so sum of neighbors < 0)  
        Do this iteration for at most max_iters or until no new points are added
        
        At the end, any point that still has pt_chosen==-5 is set to 0 
        (not chosen).  (e.g. these correspond to dry points below MHW not
        connected to the coast.)
        
    Returns: pt_chosen array of same shape as Ztopo with values 1 (where chosen)
        or 0 (where not chosen).  

    To use as a mask to plot only the chosen points, use e.g.
        Zmasked = ma.masked_array(Ztopo, mask=logical_not(pt_chosen))
        
    Sample applications:
     1. With the default parameters, choose all points with Ztopo < 0 and that 
        can be connected by a path of grid points below this elevation to 
        "deep water" where Ztopo < -5.
    """
    
    # determine if we should first select points < Z1 and march up to Z2,
    # of if we should initially select points > Z1 and march down to Z2:
    increasing = (Z1 <= Z2)
    
    if mask is None:
        mask = zeros(Ztopo.shape)
    
    if prev_pts_chosen is not None:
        # reset previous unchosen points to unset for further consideration,
        # but only in regions not masked out
        reset = logical_and(prev_pts_chosen==0, logical_not(mask))
        pt_chosen = where(reset, -5, prev_pts_chosen)
        if verbose:
            print('Resetting %i points to unset' % reset.sum())    
    else:
        # set pt_chosen=1 if Z<Z1 and =0 if Z>Z2, else =-5 (unset)
        if increasing:
            cond1 = logical_and(Ztopo < Z1, logical_not(mask))
            cond0 = logical_and(Ztopo >= Z2, logical_not(mask))
        else:
            cond1 = logical_and(Ztopo > Z1, logical_not(mask))
            cond0 = logical_and(Ztopo <= Z2, logical_not(mask))
        pt_chosen = where(cond1, 1, -5)  # mark as chosen or unknown
        pt_chosen = where(cond0, 0, pt_chosen) # mark as unchosen
    
    pt_chosen_neighbor_sum = zeros(pt_chosen.shape)
    pt_chosen_neighbor_sum[:,1:] += pt_chosen[:,:-1]
    pt_chosen_neighbor_sum[:,:-1] += pt_chosen[:,1:]
    pt_chosen_neighbor_sum[1:,:] += pt_chosen[:-1,:]
    pt_chosen_neighbor_sum[:-1,:] += pt_chosen[1:,:]

    
    # Determine points at front to use in next iteration:
    # Points that have been chosen and for which at least one neighbor is
    # unmarked and needs to be determined.
    set_next_i, set_next_j = where(logical_and(pt_chosen == 1, \
                               pt_chosen_neighbor_sum < 0))

    nfront = len(set_next_i)
    if verbose:
        print('Initially: %i cells out of %i cells on frontier' \
                % (nfront, prod(Z.shape)))

    if max_iters is None:
        max_iters = prod(Ztopo.shape)  # upper bound, should need much fewer
    
    print('Selecting points with Z1 = %g, Z2 = %g, max_iters=%i' \
          % (Z1,Z2,max_iters))

    iter_count = 0
    maxi = Ztopo.shape[0] - 1
    maxj = Ztopo.shape[1] - 1

    for k in range(max_iters):
        if len(set_next_i) == 0:
            # done
            break
        iter_count += 1
        if verbose==2:
            print('Iteration %3i: there are %7i points on frontier' \
                   % (k, len(set_next_i)))
        new_set_next_i = []  # for the next iteration
        new_set_next_j = []  # for the next iteration
        for (i,j) in zip(set_next_i,set_next_j):
            for (ii,jj) in [(i-1,j),(i+1,j),(i,j-1),(i,j+1)]:
                if (ii>=0) and (ii<=maxi) and (jj>=0) and (jj<=maxj):
                    if increasing:
                        cond1 = (Ztopo[ii,jj] < Z2)
                    else:
                        cond1 = (Ztopo[ii,jj] > Z2)
                    if (pt_chosen[ii,jj] < 0) and cond1 \
                                and (not mask[ii,jj]):
                        pt_chosen[ii,jj] = 1
                        new_set_next_i.append(ii)
                        new_set_next_j.append(jj)
                else:
                    # (ii,jj) outside grid
                    pass
        set_next_i = new_set_next_i
        set_next_j = new_set_next_j
        
            
    # any remaining unset points are dry points below Z2
    pt_chosen = where(pt_chosen<0, 0, pt_chosen)

    print('Done after %i iterations with %i points chosen' \
          % (iter_count,pt_chosen.sum()))

    return pt_chosen

    
    



from __future__ import absolute_import
from __future__ import print_function
from petsc4py import PETSc
import numpy as np
import DMPFOR
from six.moves import range



global_nx =3
global_ny =2
dof=4

da = PETSc.DA().create(dim=2,
dof=dof,
sizes=[global_nx, global_ny], 
#periodic_type = PETSc.DA.PeriodicType.GHOSTED_XYZ,
#stencil_type=self.STENCIL,
#stencil_width=2,
comm=PETSc.COMM_WORLD)


gVec = da.createGlobalVector()
lVec = da.createLocalVector()




ranges = da.getRanges()

nx_start = ranges[0][0]
nx_end = ranges[0][1]
ny_start = ranges[1][0]
ny_end = ranges[1][1]

nx = nx_end - nx_start
ny = ny_end - ny_start


q = np.empty((dof, nx, ny), order='F')

for i in range(0,nx):
    for j in range(0,ny):
        for k in range(0,dof):
            q[k,i,j] = k+10*i+100*j

gVec.array = q

q = gVec.array.reshape((dof, nx, ny), order='F')

print("da array from python")
print(q)


print("da array from fortran")
DMPFOR.dmpfor(q,dof,nx,ny)


print("da array from python after rolling axes using rollaxis")
rolled_q_1 = np.rollaxis(q,0,3)
rolled_q_1 = np.reshape(rolled_q_1,(nx,ny,dof),order='F')
print(rolled_q_1)
print("da array from fortran after rolling axes using rollaxis")
DMPFOR.dmpfor(rolled_q_1,nx,ny,dof)


print("da array from python after rolling axes using element by element copy")
rolled_q_2 = np.empty((nx,ny,dof),order='F')
for i in range(0,nx):
    for j in range(0,ny):
        for k in range(0,dof):
            rolled_q_2[i,j,k] = q[k,i,j]
print(rolled_q_2)
print("da array from fortran after rolling axes using element by element copy")
DMPFOR.dmpfor(rolled_q_2,nx,ny,dof)










#!/usr/bin/env python
# encoding: utf-8
r"""
Module containing all Pyclaw solution objects
"""

import os
import logging

from .geometry import Patch, Dimension, Domain

# ============================================================================
#  Solution Class
# ============================================================================
class Solution(object):
    r"""
    Pyclaw patch container class
        
    :Input and Output:
    
        Input and output of solution objects is handle via the io package.
        Solution contains the generic methods :meth:`write`, :meth:`read` and
        :meth:`plot` which then figure out the correct method to call.  Please
        see the io package for the particulars of each format and method and 
        the methods in this class for general input and output information.
    
    :Properties:
    
        If there is only one state and patch belonging to this solution, 
        the solution will appear to have many of the attributes assigned to its
        one state and patch.  Some parameters that have in the past been
        parameters for all patch,s are also reachable although Solution does not
        check to see if these parameters are truly universal.

        Patch Attributes:
            'dimensions'
        State Attributes:
            't','num_eqn','q','aux','capa','problem_data'
            
            
    :Initialization:
        
        The initialization of a Solution can happen one of these ways
        
            1. `args` is empty and an empty Solution is created
            2. `args` is an integer (the number of components of q), a single
               State, or a list of States and is followed
               by the appropriate :ref:`geometry <pyclaw_geometry>` object
               which can be one of:
                
                 - (:class:`~pyclaw.geometry.Domain`)
                 - (:class:`~pyclaw.geometry.Patch`) - A domain is created
                   with the patch or list of patches provided.
                 - (:class:`~pyclaw.geometry.Dimension`) - A domain and 
                   patch is created with the dimensions or list of 
                   dimensions provided.
            3. `args` is a variable number of keyword arguments that describes the
               location of a file to be read in to initialize the object
    
    :Examples:

        >>> import clawpack.pyclaw as pyclaw
        >>> x = pyclaw.Dimension('x',0.,1.,100)
        >>> domain = pyclaw.Domain((x))
        >>> state = pyclaw.State(domain,3,2)
        >>> solution = pyclaw.Solution(state,domain)
    """
    def __getattr__(self, key):
        if key in ('t','num_eqn','mp','mF','q','p','F','aux','capa',
                   'problem_data','num_aux',
                   'num_dim', 'p_centers', 'p_edges', 'c_centers', 'c_edges',
                   'num_cells', 'lower', 'upper', 'delta', 'centers', 'edges',
                   'gauges', 'num_eqn', 'num_aux', 'grid', 'problem_data'):
            return self._get_base_state_attribute(key)
        else:
            raise AttributeError("'Solution' object has no attribute '"+key+"'")

    def __setattr__(self, key, value):
        if key in ('t','mp','mF'):
            self.set_all_states(key,value)
        else:
            self.__dict__[key] = value

    # ========== Attributes ==================================================
    
    # ========== Properties ==================================================
    @property
    def state(self):
        r"""(:class:`State`) - Base state is returned"""
        return self.states[0]
    @property
    def patch(self):
        r"""(:class:`Patch`) - Base state's patch is returned"""
        return self.domain.patch

    @property
    def start_frame(self):
        r"""(int) - : Solution start frame number in case the `Solution`
        object is initialized by loading frame from file"""
        return self._start_frame
    _start_frame = 0


    # ========== Class Methods ===============================================
    def __init__(self,*arg,**kargs):
        r"""Solution Initiatlization Routine
        
        See :class:`Solution` for more info.
        """

        # select package to build solver objects from, by default this will be
        # the package that contains the module implementing the derived class
        # for example, if Solution is implemented in 'clawpack.petclaw.solution', then 
        # the computed claw_package will be 'clawpack.petclaw'

        import sys
        if 'claw_package' in list(kargs.keys()):
            claw_package = kargs['claw_package']
        else:
            claw_package = None

        if claw_package is not None and claw_package in sys.modules:
            self.claw_package = sys.modules[claw_package]
        else:
            def get_clawpack_dot_xxx(modname): return modname.rpartition('.')[0]
            claw_package_name = get_clawpack_dot_xxx(self.__module__)
            if claw_package_name in sys.modules:
                self.claw_package = sys.modules[claw_package_name]
            else:
                raise NotImplementedError("Unable to determine solver package, please provide one")

        State = self.claw_package.State

        self.states = []
        self.domain = None
        if len(arg) == 1:
            # Load frame
            frame = arg[0]
            if not isinstance(frame,int):
                raise Exception('Invalid pyclaw.Solution object initialization')
            if ('count_from_zero' in kargs):
                if (kargs['count_from_zero'] == True):
                    self._start_frame = 0
                else:
                    self._start_frame = frame
                kargs.pop('count_from_zero')
            self.read(frame,**kargs)

        elif len(arg) == 2:
            #Set domain
            if isinstance(arg[1],Domain):
                self.domain = arg[1]
            else:
                if not isinstance(arg[1],(list,tuple)):
                    arg[1] = list(arg[1])
                if isinstance(arg[1][0],Dimension):
                    self.domain = Domain(Patch(arg[1]))
                elif isinstance(arg[1][0],Patch):
                    self.domain = Domain(arg[1])
                else:
                    raise Exception("Invalid arguments for Solution initialization.")

            #Set state
            if isinstance(arg[0],State):
                # Single State
                self.states.append(arg[0])
            elif isinstance(arg[0],(list,tuple)):
                if isinstance(arg[0][0],State):
                    # List of States
                    self.states = arg[0]
                elif isinstance(arg[0][0],int):
                    self.states = State(self.domain,arg[0][0],arg[0][1])
                else:
                    raise Exception("Invalid arguments for Solution initialization.")
            elif isinstance(arg[0],int):
                self.states.append(State(self.domain,arg[0]))
            if self.states == [] or self.domain is None:
                raise Exception("Invalid arguments for Solution initialization.")
        elif len(arg) == 0:
            if 'frame' in kargs:
                frame = kargs.pop('frame')
                self.read(frame,**kargs)
            elif not kargs:
                pass  # With no arguments, initialize empty solution
            else:
                raise Exception("Invalid arguments for Solution initialization.")
        else:
            raise Exception("Invalid arguments for Solution initialization.")
                
                
    def is_valid(self):
        r"""
        Checks to see if this solution is valid
        
        The Solution checks to make sure it is valid by checking each of its
        states.  If an invalid state is found, a message is logged what
        specifically made this solution invalid.
       
        :Output:
         - (bool) - True if valid, false otherwise
        """
        return all([state.is_valid() for state in self.states])


    def __str__(self):
        output = "states:\n"
        # This is information about each of the states
        for state in self.states:
            output = output + str(state)
        return str(output)
    
    
    def set_all_states(self,attr,value,overwrite=True):
        r"""
        Sets all member states attribute 'attr' to value
        
        :Input:
         - *attr* - (string) Attribute name to be set
         - *value* - (id) Value for attribute
         - *overwrite* - (bool) Whether to overwrite the attribute if it 
           already exists.  ``default = True``
        """
        for state in self.states:
            if getattr(state,attr) is None or overwrite:
                setattr(state,attr,value) 
    
                    
    def _get_base_state_attribute(self, name):
        r"""
        Return base state attribute
        
        :Output:
         - (id) - Value of attribute from ``states[0]``
        """
        return getattr(self.states[0],name)
    
    
    def __copy__(self):
        return self.__class__(self)
    
    
    def __deepcopy__(self,memo={}):
        import copy
        # Create basic container
        result = self.__class__()
        result.__init__()
        
        # Populate the states
        for state in self.states:
            result.states.append(copy.deepcopy(state))
        result.domain = copy.deepcopy(self.domain)
        
        return result
    
    
    # ========== IO Functions ================================================
    def write(self,frame,path='./',file_format='ascii',file_prefix=None,
                write_aux=False,options={},write_p=False):
        r"""
        Write out a representation of the solution

        Writes out a suitable representation of this solution object based on
        the format requested.  The path is built from the optional path and
        file_prefix arguments.  Will raise an IOError if unsuccessful.

        :Input:
         - *frame* - (int) Frame number to append to the file output
         - *path* - (string) Root path, will try and create the path if it 
           does not already exist. ``default = './'``
         - *format* - (string or list of strings) a string or list of strings 
           containing the desired output formats. ``default = 'ascii'``
         - *file_prefix* - (string) Prefix for the file name.  Defaults to
           the particular io modules default.
         - *write_aux* - (book) Write the auxiliary array out as well if 
           present. ``default = False``
         - *options* - (dict) Dictionary of optional arguments dependent on 
           which format is being used. ``default = {}``
        """
        # Determine if we need to create the path
        path = os.path.expandvars(os.path.expanduser(path))
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError:
                print("directory already exists, ignoring")  

        # Call the correct write function based on the output format
        if isinstance(file_format,str):
            format_list = [file_format]
        elif isinstance(file_format,list):
            format_list = file_format


        # Loop over list of formats requested
        for form in format_list:
            write_func = self.get_write_func(form)

            if file_prefix is None:
                write_func(self,frame,path,write_aux=write_aux,
                            options=options,write_p=write_p)
            else:
                write_func(self,frame,path,file_prefix=file_prefix,
                                write_aux=write_aux,options=options,
                           write_p=write_p)
            msg = "Wrote out solution in format %s for time t=%s" % (form,self.t)
            logging.getLogger('pyclaw.fileio').info(msg)

        
    def read(self, frame, path='./_output', file_format=None, 
                          file_prefix='fort', read_aux=True, options={}, **kargs):
        r"""
        Reads in a Solution object from a file
        
        Reads in and initializes this Solution with the data specified.  This 
        function will raise an IOError if it was unsuccessful.  

        Any format must conform to the following call signiture and return
        True if the file has been successfully read into the given solution or
        False otherwise.  Options is a dictionary of parameters that each
        format can specify.  See the ascii module for an example.::
        
            read_<format>(solution,path,frame,file_prefix,options={})
            
        ``<format>`` is the name of the format in question.
        
        :Input:
         - *frame* - (int) Frame number to be read in
         - *path* - (string) Base path to the files to be read. 
           ``default = './_output'``
         - *file_format* - (string) Format of the file, should match on of the 
           modules inside of the io package.  ``default = None``
           but now attempts to read from header file (as of v5.9.0).
         - *file_prefix* - (string) Name prefix in front of all the files, 
           defaults to whatever the format defaults to, e.g. fort for ascii
         - *options* - (dict) Dictionary of optional arguments dependent on 
           the format being read in.  ``default = {}``
            
        :Output:
         - (bool) - True if read was successful, False otherwise
        """

        from clawpack.pyclaw.fileio.ascii import read_t
        file_format2 = None
        
        try:
            [t,num_eqn,nstates,num_aux,num_dim,num_ghost,file_format2] = \
                 read_t(frame,path,file_prefix=file_prefix)
        except:
            pass

        if file_format2 is not None:
            # value was read in from file, use it:
            file_format = file_format2

        read_func = self.get_read_func(file_format)

        if file_format == 'petsc':
            options['format'] = 'binary'
        else:
            options['format'] = file_format

        path = os.path.expandvars(os.path.expanduser(path))
        if file_prefix is None:
            read_func(self,frame,path,read_aux=read_aux,options=options)
        else:
            read_func(self,frame,path,file_prefix=file_prefix,
                                    read_aux=read_aux,options=options)
        logging.getLogger('pyclaw.fileio').info("Read in solution for time t=%s" % self.t)


    def get_read_func(self, file_format):
        if file_format[:6] == 'binary':
            # could be 'binary64' or 'binary32'
            import clawpack.pyclaw.fileio.binary
            return clawpack.pyclaw.fileio.binary.read
        elif file_format == 'ascii':
            import clawpack.pyclaw.fileio.ascii
            return clawpack.pyclaw.fileio.ascii.read
        elif file_format in ('hdf','hdf5'):
            import clawpack.pyclaw.fileio.hdf5
            return clawpack.pyclaw.fileio.hdf5.read
        elif file_format == 'petsc':
            try:
                import clawpack.petclaw.fileio
                return clawpack.petclaw.fileio.petsc.read
            except AttributeError as e:
                try:
                    from petsc4py import PETSc
                except ImportError:
                    raise ImportError("petsc4py is required for reading petsc format files, but petsc4py could not be imported.")
                raise e
        elif file_format == 'forestclaw':
            import clawpack.forestclaw.fileio.ascii
            return clawpack.forestclaw.fileio.ascii.read
        else:
            raise ValueError("File format %s not supported." % file_format)


    def get_write_func(self, file_format):
        if file_format == "forestclaw":
            import clawpack.forestclaw.fileio.ascii
            return clawpack.forestclaw.fileio.ascii.write
        elif file_format == "vtk":
            import clawpack.pyclaw.fileio.claw_vtk
            return clawpack.pyclaw.fileio.claw_vtk.write
        else:
            try:
                import importlib
                return importlib.import_module("clawpack.pyclaw.fileio.%s"
                                               % file_format).write
            except:
                raise ValueError("File format %s not found." % file_format)

        
    def plot(self):
        r"""
        Plot the solution
        """
        raise NotImplementedError("Direct solution plotting has not been " +
            "implemented as of yet, please refer to the plotting module for" +
            " how to plot solutions.")


if __name__ == "__main__":
    import doctest
    doctest.testmod()



r"""
Module containing the PetClaw solvers

This file currently only exists so that these solvers have a different
__module__ property, used by pyclaw.solver.Solver.__init__ to
determine the containing claw_package to use.
"""

from __future__ import absolute_import
from clawpack import pyclaw

class ClawSolver1D(pyclaw.ClawSolver1D):
    r"""
    Parallel solver for 1D problems using classic Clawpack algorithms.
    """

    __doc__ += pyclaw.util.add_parent_doc(pyclaw.ClawSolver1D)

class ClawSolver2D(pyclaw.ClawSolver2D):
    r"""
    Parallel solver for 2D problems using classic Clawpack algorithms.
    """

    __doc__ += pyclaw.util.add_parent_doc(pyclaw.ClawSolver2D)
    
class ClawSolver3D(pyclaw.ClawSolver3D):
    r"""
    Parallel solver for 3D problems using classic Clawpack algorithms.
    """

    __doc__ += pyclaw.util.add_parent_doc(pyclaw.ClawSolver3D)



#!/usr/bin/env python
# encoding: utf-8
r'''
Generate ClawPack environment variables for bash and csh

Checks to see which Clawpack projects/repositories have been downloaded
by either checking in the CLAW directory (which defaults to the current
directory) or by overridden paths set in the command line arguments.
'''

# ============================================================================
#  Distributed under the terms of the Berkeley Software Distribution (BSD) 
#  license
#                     http://www.opensource.org/licenses/
# ============================================================================

from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import getopt

# ============================================================================
# Parameters for the projects and git repositories

# Available git repositories to be considered
git_repos = ["classic","amrclaw","geoclaw","pyclaw","clawapps","clawutil",
             "clawpack-4.x","riemann","doc","visclaw","sharpclaw"]

# Project dependencies
project_dependencies = {"classic":["riemann","clawutil"],
                        "amrclaw":["riemann","clawutil"],
                        "geoclaw":["riemann","clawutil","amrclaw"],
                        "pyclaw":["riemann","clawutil"],
                        "clawapps":["riemann","clawutil"],
                        "clawutil":[],
                        "clawpack-4.x":[],
                        "riemann":["clawutil"],
                        "doc":[],
                        "visclaw":["clawutil"],
                        "sharpclaw":["riemann","clawutil"]}

# This was intended to be a prefered means for adding and deleting paths for
# the resulting output script but we need a csh version before this will really
# work correctly without adding a lot of complexity to this utility (KTM)
bash_path_modification_functions = """
# These are utility functions for manipulating paths
var_append () {
    # Check to see if the variable exists
    if [ -z "${1}" ]; then
        export ${1}="${2}"
    else
        var_remove $1 $2
        export ${1}="`/usr/bin/printenv $1`:${2}"
    fi
}
var_prepend () {
    # Check to see if variable exists
    if [ -z "${1}" ]; then
        export ${1}="${2}"
    else
        var_remove $1 $2
        export ${1}="${2}:`/usr/bin/printenv $1`"
    fi
}
var_remove () {
    VAR_CONTENTS=`/usr/bin/printenv $1`
    NEW_VAR=`echo -n $VAR_CONTENTS | awk -v RS=: -v ORS=: '$0 != "'$2'"' | sed 's/:$//'`
    export ${1}=${NEW_VAR}
} 

path_append () { var_append PATH $1; }
path_prepend () { var_prepend PATH $1; }
path_remove () { var_remove PATH $1; }
python_append () { var_append PYTHONPATH $1;}
python_prepend () { var_prepend PYTHONPATH $1;}
python_remove () { var_remove PYTHONPATH $1;}  

"""

# ============================================================================
#  Help display
class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

repo_options = ''
for repo in git_repos:
    repo_options += "  --%s= (string) Path to %s project.  (default == $CLAW/%s)\n" % (repo,repo,repo)
help_message = '''
Generate ClawPack environment variables for bash and csh

Checks to see which Clawpack projects/repositories have been downloaded
by either checking in the CLAW directory (which defaults to the current
directory) or by overridden paths set in the command line arguments.  The 
currently accepted project/repository names are

  %s

Command line script parameters:
  -v, --verbose - Verbose output (default == True)
  -h, --help - Display help message
  -o, --output= (string) - The base name for the output bash and csh files 
                           (default == "setenv")
  -s, --shell= (string) - Type of shell script to output, valid options include
                         'csh', 'bash', 'sh', or 'both'.  The option 'both'
                         will output both a "csh" and "sh" compatable file.
                         (default == 'both')
                           
Project path options:
  -c, --claw= (string) - Path to base CLAW directory.  If this option is choosen
                         and a path provided then all project paths will be set
                         to as this as it's base, regardless of what has been 
                         set before this flag.  If you want to have a specific
                         path to the CLAW directory, give this flag first and 
                         then specify specific paths to the projects.  
                         (default == cwd)
%s 
''' % ('\n  '.join(git_repos),repo_options)

# ============================================================================
#  Helper functions
def write_environment_variable(csh_handle,bash_handle,var,value):
    if csh_handle is not None:
        csh_handle.write('setenv %s "%s"\n' % (var.upper(),value))
    if bash_handle is not None:
        bash_handle.write('export %s="%s"\n' % (var.upper(),value))

def check_repos_dependencies(project_name,available_projects):
    r"""Checks that required repositories of project_name are present"""
    missing_projects = []
    for dependency in project_dependencies[project_name]:
        if dependency not in available_projects:
            missing_projects.append(dependency)
    if len(missing_projects) > 0:
        return missing_projects
    return None

# ============================================================================
def write_env_files(claw_path,verbose=True,outfile_base="setenv",
                                                shell_type='both',**kargs):
    # Find projects
    available_projects = {}
    print("Found the following Clawpack projects:")
    for repo in git_repos:
        if repo in list(kargs.keys()):
            if os.path.exists(kargs[repo]):
                available_projects[repo] = kargs[repo]
                if verbose:
                    print("  %s -- %s" % (repo,available_projects[repo]))
                else:
                    print("  %s" % (repo))
            else:
                raise Exception("Project %s not found at custom path %s" % (repo,kargs[repo]))
        else:
            if os.path.exists(os.path.join(claw_path,repo)):
                available_projects[repo] = os.path.join(claw_path,repo)
                if verbose:
                    print("  %s -- %s" % (repo,available_projects[repo]))
                else:
                    print("  %s" % (repo))
            else:
                if verbose:
                    print("  %s -- *** Not found ***" % repo)
    
    # Check here for environment dependencies
    for project in available_projects.keys():
        missing_projects = check_repos_dependencies(project,list(available_projects.keys()))
        if missing_projects is not None:
            error_msg = "The project %s depends on the following missing projects:" % project_name
            error_msg += ("\n %s" % name for name in missing_projects)
            
    # =========================================================================
    #  Write out out_file_base.csh and out_file_base.sh
    # Open output files
    boiler_plate = ("# Clawpack environment settings\n")
    if "csh" in shell_type:
        csh_file = open(os.path.join(claw_path,".".join((outfile_base,"csh"))),'w')
        csh_file.write(boiler_plate)
    else:
        csh_file = None
    if "bash" == shell_type or "sh" == shell_type:
        bash_file = open(os.path.join(claw_path,".".join((outfile_base,"bash"))),'w')
        bash_file.write(boiler_plate)
    else:
        bash_file = None
    
    # Write out variables
    python_path = "${PYTHONPATH}"
    matlab_path = "${MATLABPATH}"
    
    print("")
    print("The following variables will be set:")
    print("  CLAW = %s" % claw_path)
    write_environment_variable(csh_file,bash_file,"CLAW",claw_path)
    
    if "clawutil" in available_projects:
        print("  CLAWUTIL = %s" % available_projects["clawutil"])
        write_environment_variable(csh_file,bash_file,"CLAWUTIL",available_projects["clawutil"])
        python_path = ":".join((os.path.join(available_projects["clawutil"],"src","python"),python_path))
        
    if "classic" in available_projects:
        print("  CLASSICCLAW = %s" % available_projects["classic"])
        write_environment_variable(csh_file,bash_file,"CLASSICCLAW",available_projects["classic"])

    if "amrclaw" in available_projects:
        print("  AMRCLAW = %s" % available_projects["amrclaw"])
        write_environment_variable(csh_file,bash_file,"AMRCLAW",available_projects["amrclaw"])

    if "geoclaw" in available_projects:
        python_path = ":".join((os.path.join(available_projects["geoclaw"],"src","python"),python_path))
        print("  GEOCLAW = %s" % available_projects["geoclaw"])
        write_environment_variable(csh_file,bash_file,"GEOCLAW",available_projects["geoclaw"])

    if "pyclaw" in available_projects:
        python_path = ":".join((os.path.join(available_projects["pyclaw"],"src"),python_path))
        print("  PYCLAW = %s" % available_projects["pyclaw"])
        write_environment_variable(csh_file,bash_file,"PYCLAW",available_projects["pyclaw"])

    if "clawapps" in available_projects:
        print("  CLAWAPPS = %s" % available_projects["clawapps"])
        write_environment_variable(csh_file,bash_file,"CLAWAPPS",available_projects["clawapps"])
    if "doc" in available_projects:
        pass
        
    if "riemann" in available_projects:
        python_path = ":".join((os.path.join(available_projects["riemann"],"src","python"),python_path))
        print("  RIEMANN = %s" % available_projects["riemann"])
        write_environment_variable(csh_file,bash_file,"RIEMANN",available_projects["riemann"])

    if "visclaw" in available_projects:
        python_path = ":".join((os.path.join(available_projects["visclaw"],"src","python"),python_path))
        matlab_path = ":".join((os.path.join(available_projects["visclaw"],"src","matlab"),matlab_path))
        print("  VISCLAW = %s" % available_projects["visclaw"])
        write_environment_variable(csh_file,bash_file,"VISCLAW",available_projects["visclaw"])

    if "shaprclaw" in available_projects:
        python_path = ":".join((os.path.join(available_projects["sharpclaw"],"src","python"),python_path))
        print("  SHARPCLAW = %s" % available_projects["sharpclaw"])
        write_environment_variable(csh_file,bash_file,"SHARPCLAW",available_projects["sharpclaw"])

    if "clawpack-4.x" in available_projects:
        # python_path = ":".join((os.path.join(available_projects["clawpack-4.x"],"python"),python_path))
        print("  CLAW_4 = %s" % available_projects["clawpack-4.x"])
        write_environment_variable(csh_file,bash_file,"CLAW_4",available_projects["clawpack-4.x"])

    if len(python_path) > 13:
        print("  PYTHONPATH = %s" % python_path)
        write_environment_variable(csh_file,bash_file,"PYTHONPATH",python_path)
    if len(matlab_path) > 13:
        print("  MATLABPATH = %s" % matlab_path)
        write_environment_variable(csh_file,bash_file,"MATLABPATH",matlab_path)
    print("")
        
    # Close output files
    if csh_file is not None:
        csh_file.close()
    if bash_file is not None:
        bash_file.close()

if __name__ == "__main__":    
    # Parse input arguments
    argv = sys.argv
    project_paths = {}
    try:
        try:
            long_options = ["help","output=","verbose","shell=",
                 "claw="]
            for proj_name in git_repos:
                long_options.append("%s=" % proj_name)
            opts, args = getopt.getopt(argv[1:], "ho:vs:c",long_options)
        except getopt.error as msg:
            raise Usage(msg)
            
        # Default script parameter values
        verbose = False
        out_file_base = "setenv"
        shell_type = 'both'
        
        # Default claw path
        claw_path = os.path.abspath(os.curdir)
    
        # option processing
        for option, value in opts:
            # Script parameters
            if option in ("-v","--verbose"):
                verbose = True
            if option in ("-o","--output"):
                out_file_base = value
            if option in ("-s","--shell"):
                shell_type = value
            if option in ("-h","--help"):
                raise Usage(help_message)
                                
            # Project path overrides
            if option in ("-c","--claw"):
                claw_path = os.path.abspath(value)
            for proj_name in git_repos:
                if option == "--%s" % proj_name:
                    project_paths[proj_name] = os.path.abspath(value)
                            
    except Usage as err:
        print(sys.argv[0].split("/")[-1] + ": " + str(err.msg), file=sys.stderr)
        print("\t for help use --help", file=sys.stderr)
        sys.exit(2)

    sys.exit(write_env_files(claw_path,verbose=verbose,
                             outfile_base=out_file_base,shell_type=shell_type,
                             **project_paths))
                



#!/usr/bin/env python

"""Base AMRClaw data class for writing out data parameter files."""

import os

import clawpack.clawutil.data

class AmrclawInputData(clawpack.clawutil.data.ClawData):
    r"""
    Data object containing AMRClaw input data.

    Extends ClawInputData adding necessary data for AMR.
    """
    def __init__(self, clawdata):

        # Some defaults are inherited from ClawInputData:
        super(AmrclawInputData,self).__init__()

        # Need to have a pointer to this so we can get num_dim and num_aux
        self._clawdata = clawdata
        
        # set max1d, maximum size of patch in each dimension
        # and memsize, initial length of alloc array 
        #    use default values so that repeated doubling gets as close as
        #    possible to 2**30-1 = 2147483647, beyond which integer(kind=4)
        #    indexes overflow and it would be necessary to use kind=8.
        if self._clawdata.num_dim == 3:
            self.add_attribute('memsize',8388607)
            self.add_attribute('max1d',32)
        elif self._clawdata.num_dim == 2:
            self.add_attribute('memsize',4194303)
            self.add_attribute('max1d',60)
        else:
            self.add_attribute('memsize',1048575)
            self.add_attribute('max1d',500)

        # Refinement control
        self.add_attribute('amr_levels_max',1)
        self.add_attribute('refinement_ratios_x',[1])
        if self._clawdata.num_dim >= 2:
            self.add_attribute('refinement_ratios_y',[1])
        if self._clawdata.num_dim == 3:
            self.add_attribute('refinement_ratios_z',[1])
        self.add_attribute('variable_dt_refinement_ratios',False)

        self.add_attribute('refinement_ratios_t',[1])
        self.add_attribute('aux_type',[])

        # Flagging control
        self.add_attribute('flag_richardson',False)
        self.add_attribute('flag_richardson_tol',-1.0)
        self.add_attribute('flag2refine',True)
        self.add_attribute('flag2refine_tol',0.05)
        self.add_attribute('regrid_interval',2)
        self.add_attribute('regrid_buffer_width',3)
        self.add_attribute('clustering_cutoff',0.7)
        self.add_attribute('verbosity_regrid',3)

        
        # debugging flags:
        self.add_attribute('dprint',False)
        self.add_attribute('eprint',False)
        self.add_attribute('edebug',False)
        self.add_attribute('gprint',False)
        self.add_attribute('nprint',False)
        self.add_attribute('pprint',False)
        self.add_attribute('rprint',False)
        self.add_attribute('sprint',False)
        self.add_attribute('tprint',False)
        self.add_attribute('uprint',False)


    def write(self, out_file='amr.data', data_source='setrun.py'):

        self.open_data_file(out_file, data_source)
    
        self.data_write('memsize')
        self.data_write('max1d')
        self.data_write('amr_levels_max')

        num_ratios = max(abs(self.amr_levels_max)-1, 1)
        if len(self.refinement_ratios_x) < num_ratios:
            raise ValueError("*** Error in data parameter: " + \
                  "require len(refinement_ratios_x) >= %s " % num_ratios)
        self.data_write('refinement_ratios_x')
        if self._clawdata.num_dim >= 2:
            if len(self.refinement_ratios_y) < num_ratios:
                raise ValueError("*** Error in data parameter: " + \
                      "require len(refinement_ratios_y) >= %s " % num_ratios)
            self.data_write('refinement_ratios_y')
        if self._clawdata.num_dim == 3:
            if len(self.refinement_ratios_z) < num_ratios:
                    raise ValueError("*** Error in data parameter: " + \
                      "require len(refinement_ratios_z) >= %s " % num_ratios)
            self.data_write('refinement_ratios_z')
        if len(self.refinement_ratios_t) < num_ratios:
            raise ValueError("*** Error in data parameter: " + \
                  "require len(refinement_ratios_t) >= %s " % num_ratios)
        self.data_write('refinement_ratios_t')
        self.data_write()  # writes blank line

        if len(self.aux_type) < self._clawdata.num_aux:
            raise ValueError("*** length of aux_type array should be " +\
                  "same as num_aux = %i" % self._clawdata.num_aux)
        if self._clawdata.num_aux > 0:
            self.data_write('aux_type')
        self.data_write()

        self.data_write('flag_richardson')
        if self.flag_richardson == False:
            # Still need to add flag_richardson to fortran, for now this works:
            self.flag_richardson_tol = -1.   
        self.data_write('flag_richardson_tol')
        self.data_write('flag2refine')
        self.data_write('flag2refine_tol')
        self.data_write('regrid_interval')
        self.data_write('regrid_buffer_width')
        self.data_write('clustering_cutoff')
        self.data_write('verbosity_regrid')
        self.data_write()

        self.data_write()

        self.data_write('dprint')
        self.data_write('eprint')
        self.data_write('edebug')
        self.data_write('gprint')
        self.data_write('nprint')
        self.data_write('pprint')
        self.data_write('rprint')
        self.data_write('sprint')
        self.data_write('tprint')
        self.data_write('uprint')
        self.data_write()

        self.close_data_file()


# ==============================================================================
#  Region data object
class RegionData(clawpack.clawutil.data.ClawData):
    r""""""

    def __init__(self,regions=None,num_dim=None):

        super(RegionData,self).__init__()

        if regions is None or not isinstance(regions,list):
            self.add_attribute('regions',[])
        else:
            self.add_attribute('regions',regions)
        self.add_attribute('num_dim',num_dim)


    def write(self,out_file='regions.data',data_source='setrun.py'):


        self.open_data_file(out_file,data_source)

        self.data_write(value=len(self.regions),alt_name='num_regions')
        for regions in self.regions:
            # write minlevel,maxlevel as integers, remaining floats in e format:
            self._out_file.write("%i  %i " % (regions[0], regions[1]))
            self._out_file.write((len(regions)-2)*"%20.14e  " % tuple(regions[2:]) +"\n")
            
        self.close_data_file()


    def read(self, path, force=False):
        r"""Read in region data from file at path
        """

        data_file = open(os.path.abspath(path),'r')

        # Read past comments and blank lines
        header_lines = 0
        ignore_lines = True
        while ignore_lines:
            line = data_file.readline()
            if line[0] == "#" or len(line.strip()) == 0:
                header_lines += 1
            else:
                break

        # Read in number of regions        
        num_regions, tail = line.split("=:")
        num_regions = int(num_regions)
        varname = tail.split()[0]
        if varname != "num_regions":
            raise IOError("It does not appear that this file contains region data.")

        # Read in each region
        self.regions = []
        for n in range(num_regions):
            line = data_file.readline().split()
            self.regions.append([int(line[0]), int(line[1])] + [float(a) for a in line[2:]])

        data_file.close()


# ==============================================================================
#  FlagRegion object (new in v5.7.0)

class FlagRegion(clawpack.clawutil.data.ClawData):
    
    def __init__(self, num_dim, region=None):
        r"""
        Note: for backward compatibility, can pass in a old-style region 
        in the form (in 2d):
            [minlevel,maxlevel,t1,t2,x1,x2,y1,y2]
        and it will be converted to the new form.
        
        """
        
        super(FlagRegion,self).__init__()
        self.add_attribute('name','')
        self.add_attribute('num_dim',num_dim)
        self.add_attribute('minlevel',None)
        self.add_attribute('maxlevel',None)
        self.add_attribute('t1',None)
        self.add_attribute('t2',None)
        self.add_attribute('spatial_region_type',None)
        self.add_attribute('spatial_region',None)
        self.add_attribute('spatial_region_file',None)
        
        if region is not None:
            self.convert_old_region(region)

        
    def convert_old_region(self, region):
        """
        Take a list region = [minlevel, maxlevel, t1, t2, x1, x2, y1, y2]
        in the old style and convert to a new flagregion.
        """
        self.minlevel = region[0]
        self.maxlevel = region[1]
        self.t1 = region[2]
        self.t2 = region[3]
        self.spatial_region_type = 1  # rectangle
        self.spatial_region = region[4:]
        
    def read_spatial_region(self, fname=None):
        """
        Reads a ruled rectangle and assigns to self.spatial_region
        But this isn't really supported yet, since cannot write to 
        flagregions.data.
        """
        from clawpack.amrclaw import region_tools
        
        if fname is None:
            fname = self.spatial_region_file
        if not os.path.isfile(fname):
            print('*** Could not find file ',fname)
        else:
            self.spatial_region = region_tools.RuledRectangle(fname)
            
        
# ==============================================================================
#  FlagRegionData object  (new in v5.7.0)

class FlagRegionData(clawpack.clawutil.data.ClawData):
    r"""To replace RegionData, allowing more flexibility."""

    def __init__(self,flagregions=None,num_dim=2):

        super(FlagRegionData,self).__init__()

        if flagregions is None or not isinstance(flagregions,list):
            self.add_attribute('flagregions',[])
        else:
            self.add_attribute('flagregions',flagregions)
        self.add_attribute('num_dim',num_dim)


    def write(self,out_file='flagregions.data',data_source='setrun.py'):
        """
        Write out the flagregion.data file that will be read by the Fortran
        code.  
        """

        self.open_data_file(out_file,data_source)

        self.data_write(value=len(self.flagregions),alt_name='num_flagregions')
        for flagregion in self.flagregions:
            flagregion._out_file = self._out_file
            # write minlevel,maxlevel as integers, t1,t2 as floats:
            flagregion.data_write()
            flagregion.data_write('name')
            flagregion.data_write('minlevel')
            flagregion.data_write('maxlevel')
            flagregion.data_write('t1')
            flagregion.data_write('t2')
            flagregion.data_write('spatial_region_type')
            
            if flagregion.spatial_region_type == 1:
                # spatial_region is an extent [x1,x2,y1,y2]
                assert len(flagregion.spatial_region) == 4, \
                    "*** FlagRegionData.write requires len(spatial_region) = 4"
                flagregion.data_write('spatial_region')
                
            elif flagregion.spatial_region_type == 2:
                if type(flagregion.spatial_region_file) is str:
                    # assumed to be path to RuledRectangle file
                    flagregion.data_write('spatial_region_file')
                else:
                    raise ValueError('*** FlagRegionData.write requires path to RuledRectangle')
                    
        self.close_data_file()


    def read(self, path='flagregions.data', force=False):
        r"""Read in flagregion data from file at path
        """

        with open(os.path.abspath(path),'r') as data_file:
            # Read past comments and blank lines
            header_lines = 0
            ignore_lines = True
            while ignore_lines:
                line = data_file.readline()
                if line[0] == "#" or len(line.strip()) == 0:
                    header_lines += 1
                else:
                    break
    
            # Read in number of flagregions        
            num_flagregions, tail = line.split("=:")
            num_flagregions = int(num_flagregions)
            varname = tail.split()[0]
            if varname != "num_flagregions":
                raise IOError("It does not appear that this file contains flagregion data.")
    
            print('Will read %i flagregions' % num_flagregions)
            
            # Read in each flagregion
            self.flagregions = []
            for n in range(num_flagregions):
                flagregion = FlagRegion(num_dim=2)
                line = data_file.readline() # blank
                line = data_file.readline().split()
                flagregion.name = str(line[0])[1:-1] # strip extra quotes
                line = data_file.readline().split()
                flagregion.minlevel = int(line[0])
                line = data_file.readline().split()
                flagregion.maxlevel = int(line[0])
                line = data_file.readline().split()
                flagregion.t1 = float(line[0])
                line = data_file.readline().split()
                flagregion.t2 = float(line[0])
                line = data_file.readline().split()
                flagregion.spatial_region_type = int(line[0])
                
                line = data_file.readline().split()
                if flagregion.spatial_region_type == 1:
                    flagregion.spatial_region = [float(val) for val in line[:4]]
                elif flagregion.spatial_region_type == 2:
                    # the next line is assumed to be path to a
                    # RuledRectangle file, strip quotes
                    flagregion.spatial_region_file = line[0][1:-1] 
                    # read the file: 
                    flagregion.read_spatial_region()
                else:
                    raise ValueError('*** Unrecognized spatial_region_type')
                
                self.flagregions.append(flagregion)
    
# ==============================================================================
#  Gauge data object
class GaugeData(clawpack.clawutil.data.ClawData):
    r""""""

    defaults = {"file_format":"ascii", "display_format":"e15.7",
                "q_out_fields":"all", "aux_out_fields":"none",
                "min_time_increment":0.0, "gtype":"stationary"}

    @property
    def gauge_numbers(self):
        if len(self.gauges) == 1:
            return [self.gauges[0][0]]
        else:
            return [gauge[0] for gauge in self.gauges]


    def __init__(self, num_dim=None):
        # num_dim is an argument for backward compatibility, but no longer used
        super(GaugeData,self).__init__()
        self.add_attribute('gauges',[])

        for (value, default) in self.defaults.items():
            self.add_attribute(value, default)


    def __str__(self):
        output = "Gauges: %s\n" % len(self.gauges)
        for gauge in self.gauges:
            output = "\t".join((output,"%4i:" % gauge[0]))
            output = " ".join((output,"%19.10e" % gauge[1]))
            output = " ".join((output,"%17.10e" % gauge[2]))
            output = " ".join((output,"%13.6e" % gauge[3]))
            output = " ".join((output,"%13.6e\n" % gauge[4]))
        output = "\n\n".join((output, "Output Format: %s\n" % self.file_format))
        output = "\n\n".join((output, "Gauge type: %s\n" % self.gtype))
        output = "\t".join((output, "display: %s" % self.display_format))
        output = "\n\t".join((output, "q fields: %s" % self.q_out_fields))
        output = "\n\t".join((output, "aux fields: %s" % self.aux_out_fields))
        output = "\n\t".join((output, "min. time increment: %s" 
                                                     % self.min_time_increment))
        return output


    def write(self, num_eqn, num_aux, out_file='gauges.data', 
                                      data_source='setrun.py'):
        r"""Write out gauge information data file."""

        # Check to make sure we have only unique gauge numbers
        if len(self.gauges) > 0:
            if len(self.gauge_numbers) != len(set(self.gauge_numbers)):
                raise Exception("Non unique gauge numbers specified.")

        for gaugeno in self.gauge_numbers:
            if abs(gaugeno) >= 2**31:
                raise Exception("Gauge number %i is too large, must be < 2**31"\
                    % gaugeno)


        # Write out gauge data file
        self.open_data_file(out_file,data_source)
        self.data_write(name='ngauges',value=len(self.gauges))
        for gauge in self.gauges:
            format = "%4i" + (len(gauge)-3) * "  %17.10e" + 2 * "  %13.6e" + "\n"
            self._out_file.write(format % tuple(gauge))
        self.data_write()
        
        # Expand all gauge format option dictionaries
        for key in self.defaults.keys():
            self.expand_gauge_format_option(key)

        # File format
        self._out_file.write("# File format (1=ascii, 2=binary64, 3=binary32)\n")
        format_map = {'ascii':1, 'binary':2, 'binary64':2, 'binary32':3}
        for gauge_num in self.gauge_numbers:
            try:
                file_format = format_map[self.file_format[gauge_num].lower()]
            except KeyError:
                raise ValueError("Invalid file format %s requested." % self.file_format[gauge_num])
            self._out_file.write("%s " % file_format)
        self._out_file.write("\n")
        self.data_write()

        # Display format for each gauge
        self._out_file.write("# Display format (for ascii files)\n")
        for gauge_num in self.gauge_numbers:
            self._out_file.write("%s " % self.display_format[gauge_num])
        self._out_file.write("\n\n")

        # Minimum time increment output
        self._out_file.write("# Minimum output increment\n")
        for gauge_num in self.gauge_numbers:
            self._out_file.write("%s " % self.min_time_increment[gauge_num])
        self._out_file.write("\n\n")

        # Gauge type
        self._out_file.write("# Gauge type (1=stationary, 2=Lagrangian)\n")
        gtype_map = {'stationary':1, 'lagrangian': 2}
        for gauge_num in self.gauge_numbers:
            try:
                gtype = gtype_map[self.gtype[gauge_num].lower()]
            except KeyError:
                raise ValueError("Invalid gauge type %s requested." \
                        % self.gtype[gauge_num])
            self._out_file.write("%s " % gtype)
        self._out_file.write("\n")
        self.data_write()

        # Which q fields to output
        self._out_file.write("# q fields\n")
        for gauge_num in self.gauge_numbers:
            # Handle special values of "all" and "none"
            if isinstance(self.q_out_fields[gauge_num], str):
                if self.q_out_fields[gauge_num].lower() == 'all':
                    self._out_file.write("%s\n" % " ".join(['True'] * num_eqn))
                elif self.q_out_fields[gauge_num].lower() == 'none': 
                    self._out_file.write("%s\n" % " ".join(['False'] * num_eqn))
                else:
                    raise ValueError("Unknown q field string specified, '%s'" 
                                                 % self.q_out_fields[gauge_num])
            else:
                # Specify by field number
                if not isinstance(self.q_out_fields[gauge_num], list):
                    self.q_out_fields[gauge_num] = [self.q_out_fields[gauge_num]]
                bool_list = [n in self.q_out_fields[gauge_num] 
                                                       for n in range(num_eqn)]
                bool_list = [str(value) for value in bool_list]
                self._out_file.write("%s\n" % (" ".join(bool_list)))
        self.data_write()

        # Which aux fields to output
        if num_aux > 0:
            self._out_file.write("# aux fields\n")
            for gauge_num in self.gauge_numbers:
                # Handle special values of "all" and "none"
                if isinstance(self.aux_out_fields[gauge_num], str):
                    if self.aux_out_fields[gauge_num].lower() == 'all':
                        self._out_file.write("%s\n" % " ".join(['True'] * num_aux))
                    elif self.aux_out_fields[gauge_num].lower() == 'none': 
                        self._out_file.write("%s\n" % " ".join(['False'] * num_aux))
                    else:
                        raise ValueError("Unknown q field string specified, '%s'" 
                                                     % self.aux_out_fields[gauge_num])
                else:
                    # Specify by field number
                    if not isinstance(self.aux_out_fields[gauge_num], list):
                        self.aux_out_fields[gauge_num] = [self.aux_out_fields[gauge_num]]
                    bool_list = [n in self.aux_out_fields[gauge_num] 
                                                           for n in range(num_aux)]
                    bool_list = [str(value) for value in bool_list]
                    self._out_file.write("%s\n" % (" ".join(bool_list)))

        self.close_data_file()


    def expand_gauge_format_option(self, param_name):
        r"""Construct the full gauge output specification for *param_name*

        Also handles if each *param_name* is set to a single value (not a dict)
        and then assumes that all gauges should have this parameter.  For
        example if a user set

        > file_format = 'ascii'

        this would set all gauges to have the 'ascii' format.

        """

        default_param = self.defaults[param_name]
        if not isinstance(getattr(self, param_name), dict):
            # Convert into dict for file
            default_param = getattr(self, param_name)
            setattr(self, param_name, {})

        # Check to make sure that gauges listed are actually gauges
        for gauge_num in getattr(self, param_name).keys():
            if gauge_num not in self.gauge_numbers:
                raise ValueError("Gauge number listed in format option not a ",
                                 "gauge.  gauge_id = %s" % gauge_num)

        for gauge_num in self.gauge_numbers:
            getattr(self, param_name).setdefault(gauge_num, default_param)


    def read(self, data_path="./", file_name='gauges.data'):
        r"""Read gauge data file"""
        path = os.path.join(data_path, file_name)
        gauge_file = open(path,'r')

        # Read past comments and blank lines
        header_lines = 0
        ignore_lines = True
        while ignore_lines:
            line = gauge_file.readline()
            if line[0] == "#" or len(line.strip()) == 0:
                header_lines += 1
            else:
                break

        # Read number of gauges, should be line that was last read in
        num_gauges = int(line.split()[0])

        # Read in each gauge line
        for n in range(num_gauges):
            line = gauge_file.readline().split()
            self.gauges.append([int(line[0])] + [float(a) for a in line[1:]])

        # TODO:  Read in format data

        gauge_file.close()


#  Gauge data objects
# ==============================================================================


# ==============================================================================
#  Adjoint data object
class AdjointData(clawpack.clawutil.data.ClawData):
    r""""""

    def __init__(self, use_adjoint=False, num_dim=None):

        super(AdjointData,self).__init__()

        self.add_attribute('use_adjoint',use_adjoint)
        self.add_attribute('adjoint_outdir','')
        self.add_attribute('adjoint_files',[])
        self.add_attribute('numadjoints',0)
        self.add_attribute('t1',None)
        self.add_attribute('t2',None)
        self.add_attribute('innerprod_index', None)


    def write(self,out_file='adjoint.data',data_source='setrun.py'):

        self.open_data_file(out_file,data_source)

        self.data_write('use_adjoint')
        self.data_write('adjoint_outdir')
        self.data_write('t1')
        self.data_write('t2')
        self.data_write('innerprod_index')

        self.set_adjoint_files()
        self.data_write('numadjoints')
            
        for file in self.adjoint_files:
            # if path is relative in setrun, assume it's relative to the
            # same directory that out_file comes from
            fname = os.path.abspath(os.path.join(os.path.dirname(out_file),file))
            self._out_file.write("'%s'\n" % fname)
        self.close_data_file()

    def set_adjoint_files(self):
        import glob
        from clawpack.pyclaw.fileio.ascii import read_t

        self.adjoint_files = []
        if self.use_adjoint:
            files = glob.glob(os.path.join(self.adjoint_outdir,"fort.b*"))
            files.sort()
            for file in files:
                frameno = int(file[-4:])
                #print('+++ file, frameno: ',file, frameno)
                [t,num_eqn,nstates,num_aux,num_dim,num_ghost,file_format] \
                    = read_t(frameno, self.adjoint_outdir)
                if t <= self.t2:
                    #print('+++   using this file')
                    self.adjoint_files.append(file)
            self.numadjoints = len(self.adjoint_files)
            if (len(self.adjoint_files) == 0):
                print("*** WARNING: No binary files found for adjoint output!")


    def read_adjoint_files(self):
        self.set_adjoint_files()
        # read them in for Python processing...  add if we need it



if __name__ == '__main__':
    raise Exception("Not unit tests have been defined for this module.")





""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
""" 

from __future__ import absolute_import
import numpy as np

#--------------------------
def setplot(plotdata=None):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of clawpack.visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 


    from clawpack.visclaw import colormaps

    if plotdata is None:
        from clawpack.visclaw.data import ClawPlotData
        plotdata = ClawPlotData()


    plotdata.clearfigures()  # clear any old figures,axes,items data
    

    # Figure for pcolor
    plotfigure = plotdata.new_plotfigure(name='q[0]', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.title = 'q[0]'
    plotaxes.scaled = True

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = 0
    plotitem.pcolor_cmap = colormaps.yellow_red_blue
    plotitem.pcolor_cmin = 0.1
    plotitem.pcolor_cmax = 1.0
    plotitem.add_colorbar = True
    plotitem.show = True       # show on plot?
    

    # Figure for contour
    plotfigure = plotdata.new_plotfigure(name='contour', figno=1)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.title = 'q[0]'
    plotaxes.scaled = True
    plotaxes.afteraxes = addgauges

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
    plotitem.plot_var = 0
    plotitem.contour_levels = np.linspace(0.2,0.9,8)
    plotitem.amr_contour_colors = ['b','k','r']
    plotitem.amr_celledges_show = [0,0,0]
    plotitem.patchedges_show = 0
    

    # Figure for grid cells
    plotfigure = plotdata.new_plotfigure(name='cells', figno=2)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [0,1]
    plotaxes.ylimits = [0,1]
    plotaxes.title = 'Grid patches'
    plotaxes.scaled = True

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_patch')
    plotitem.amr_patch_bgcolor = ['#ffeeee', '#eeeeff', '#eeffee']
    plotitem.amr_celledges_show = [1,0]
    plotitem.amr_patchedges_show = [1]


    #-----------------------------------------
    # Figures for gauges
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='q', figno=300, \
                    type='each_gauge')
    plotfigure.clf_each_gauge = True

    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.title = 'q'
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 0
    plotitem.plotstyle = 'b-'


    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via clawpack.visclaw.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.html_movie = 'JSAnimation'      # new style, or "4.x" for old style
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?
    plotdata.parallel = True                 # make multiple frame png's at once

    return plotdata

    
# To plot gauge locations on pcolor or contour plot, use this as
# an afteraxis function:

def addgauges(current_data):
    from clawpack.visclaw import gaugetools
    gaugetools.plot_gauge_locations(current_data.plotdata, \
         gaugenos='all', format_string='ko', add_labels=True)



r"""
Execute nosetests in all subdirectories, to run a series of quick
regression tests.

Sends output and result/errors to separate files to simplify checking
results and looking for errors.
"""

from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import tempfile
import subprocess
import unittest
import shutil
import inspect
import time

import numpy

import clawpack.geoclaw.util
import clawpack.pyclaw.solution as solution

# Support for WIP decorator
from functools import wraps
from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest

# Work in progress test decorator
def fail(message):
    raise SkipTest(message)
 
def wip(f):
    @wraps(f)
    def run_test(*args, **kwargs):
        try:
            # Set to success so we don't save out the output when we know things
            # are awry
            args[0].success = True
            f(*args, **kwargs)
        except Exception as e:
            raise SkipTest("WIP test failed: " + str(e))
        fail("test passed but marked as work in progress")
    
    return attr('wip')(run_test)


class ClawpackRegressionTest(unittest.TestCase):

    r"""Base Clawpcak regression test setup

    All regression tests for Clawpack are derived from this base class.  The
    class conforms to the *unittest* modules standards including *setUp*, 
    *runTest* and *tearDown* methods which implement various parts of the 
    regression test.  Generally these methods are responsible for the following:

     - *setUp*: Creates the temprorary directory that will house test output and
       data.   Also instantiates catching of output to both *stdout* and 
       *stderr*.  Finally, this method also calls *build_executable* which will
       call the local Makefile's build process via *make .exe*.
     - *runTest*: Actually runs the test calling the following functions by
       default:
        - *load_rundata(): Creates the *rundata* objects via the local 
          *setrun.py* file.  The resulting *rundata* object is stored in the 
          class attribute *rundata*.
        - *write_rundata_objects*: Writes out all the data objects found in the
          *rundata* class attribute.
        - *run_code*: Runs the simulation based on the *runclaw.py* script 
          (what *make output* usually calls) in a subprocess to preserve 
          parallelism.  Note that all output is redirected to the class 
          attributes *stderr* and *stdout*.
        - *check_gauges*: Default test check provided by this class.  Simply 
          checks the gauges recorded from the simulation that they are equal in
          sum to the default test data.
     - *tearDown*:  Closes output redirection and tests for success via the
       class attribute of a similar name.  If the tests were not successful the
       temporary directory contents is copied to the local directory.  The 
       temporary directory is removed at this point.


    """

    def __init__(self, methodName="runTest"):

        super(ClawpackRegressionTest, self).__init__(methodName=methodName)

        self.stdout = None
        self.stderr = None
        self.success = False
        self.remote_files = []
        self.temp_path = None
        self.test_path = os.path.dirname(inspect.getfile(self.__class__))
        if len(self.test_path) == 0:
             self.test_path = "./"
        self.rundata = None
        self.executable_name = None


    def get_remote_file(self, url, **kwargs):
        r"""Fetch file located at *url* and store in object's *temp_path*.

        Will check downloaded file's suffix to see if the file needs to be
        un-archived.

        """

        if self.temp_path is None:
            raise ValueError("Temporary data directory has not been ",
                             "set yet.  Try calling super before attempting ",
                             "to get remote files.")

        output_path = clawpack.clawutil.data.get_remote_file(url, 
                                                      output_dir=self.temp_path, 
                                                      **kwargs)
        self.remote_files.append(output_path)
        return output_path


    def setUp(self):
        r"""Create temp dir for data and setup log files.

        """

        self.temp_path = tempfile.mkdtemp()

        self.stdout = open(os.path.join(self.temp_path, "run_output.txt"), "w")
        self.stdout.write("Output from Test %s\n" % self.__class__.__name__)
        # TODO - Should change this to use the time module's formatting 
        # apparatus
        tm = time.localtime()
        year = str(tm[0]).zfill(4)
        month = str(tm[1]).zfill(2)
        day = str(tm[2]).zfill(2)
        hour = str(tm[3]).zfill(2)
        minute = str(tm[4]).zfill(2)
        second = str(tm[5]).zfill(2)
        date = 'Started %s/%s/%s-%s:%s.%s\n' % (year,month,day,hour,minute,second)
        self.stdout.write(date)
        self.stdout.write(("="*80 + "\n"))

        self.stderr = open(os.path.join(self.temp_path, "error_output.txt"), "w")
        self.stderr.write("Errors from Test %s\n" % self.__class__.__name__)
        self.stderr.write(date)
        self.stderr.write(("="*80 + "\n"))

        self.stdout.flush()
        self.stderr.flush()

        self.stdout.write("Paths:")
        self.stdout.write("  %s" % self.temp_path)
        self.stdout.write("  %s" % self.test_path)
        self.stdout.flush()
        self.build_executable()


    def build_executable(self, executable_name="xclaw"):
        r"""Build executable by running `make .exe` in test directory.

        Moves the resulting executable to the temporary directory.


        """

        try:
            self.stdout.write("Test path and class info:\n")
            self.stdout.write("  class: %s\n" % str(self.__class__))
            self.stdout.write("  class file: %s\n" % str(inspect.getfile(self.__class__)))
            self.stdout.write("  test path: %s\n" % str(self.test_path))
            self.stdout.write("  temp path: %s\n" % str(self.temp_path))
            subprocess.check_call("cd %s ; make .exe" % self.test_path, 
                                                        stdout=self.stdout,
                                                        stderr=self.stderr,
                                                        shell=True)
        except subprocess.CalledProcessError as e:
            self.tearDown()
            raise e

        self.executable_name = executable_name
        shutil.move(os.path.join(self.test_path, self.executable_name),  
                    self.temp_path)


    def load_rundata(self):
        r"""(Re)load setrun module and create *rundata* object


        """

        if 'setrun' in sys.modules:
            del(sys.modules['setrun'])
        sys.path.insert(0, self.test_path)
        import setrun
        self.rundata = setrun.setrun()
        sys.path.pop(0)


    def write_rundata_objects(self, path=None):
        r"""Write out data in the *rundata* object to *path*

        Defaults to the temporary directory path *temp_path*.

        """

        if path is None:
            path = self.temp_path

        orig_path = os.getcwd()
        os.chdir(path)
        self.rundata.write()
        os.chdir(orig_path)


    def run_code(self):
        r"""Run test code given an already compiled executable"""

        runclaw_cmd = " ".join((
                            "cd %s ;" % self.temp_path,
                            "python",
                            "$CLAW/clawutil/src/python/clawutil/runclaw.py",
                            self.executable_name,
                            self.temp_path,
                            "True",
                            "False",
                            self.temp_path))
        subprocess.check_call(runclaw_cmd, stdout=self.stdout, 
                                           stderr=self.stderr,
                                           shell=True)
        self.stdout.flush()
        self.stderr.flush()


    def runTest(self, save=False, indices=(2, 3)):
        r"""Basic run test functionality

        Note that this stub really only runs the code and performs no tests.

        :Input:
         - *save* (bool) - If *True* will save the output from this test to 
           the file *regresion_data.txt*.  Passed to *check_gauges*.  Default is
           *False*.
         - *indices* (tuple) - Contains indices to compare in the gague 
           comparison and passed to *check_gauges*.  Defaults to *(2, 3)*.

        """

        # Write out data files
        self.load_rundata()
        self.write_rundata_objects()

        # Run code
        self.run_code()

        # Perform tests
        # Override this class to perform data checks, as is this class will
        # simply check that a test runs to completion

        # If we have gotten here then we do not need to copy the run results
        self.success = True

    
    def check_frame(self, save=False, indices=[0], frame_num=1,
                          regression_data_path="regression_data.txt",
                          tolerance=1e-14):
        r"""Compare choosen frame to the comparison data

        :Input:
         - *save* (bool) - If *True* will save the output from this test to 
           the file *regresion_data.txt*.  Default is *False*.
         - *indices* (tuple) - Contains indices to compare in the gague 
           comparison.  Defaults to *(0)*.
         - *frame_num* (int) - Frame number to load from the run data.  Defaults
           to *1*. 
         - *regression_data_path* (path) - Path to the regression test data.
           Defaults to 'regression_data.txt'.
         - *tolerance* (float) - Tolerance used in the comparison, default is
           *1e-14*.
        """

        # Load test data
        data = solution.Solution(frame_num, path=self.temp_path)
        data_sum = []
        for index in indices:
            data_sum.append(data.q[index, ...].sum())
        
        # Get (and save) regression comparison data
        regression_data_file = os.path.join(self.test_path,
                                            regression_data_path)
        if save:
            numpy.savetxt(regression_data_file, data_sum)
        regression_sum = numpy.loadtxt(regression_data_file)
        # regression_sum = []
        # for index in indices:
        #     regression_sum.append(regression_data[index, :].sum())

        assert numpy.allclose(data_sum, regression_sum, tolerance), \
            "\n  new_data: %s, \n  expected: %s"  % (data_sum, regression_sum)


    def check_gauges(self, save=False, 
                           regression_data_path="regression_data.txt",
                           tolerance=1e-14):
        r"""Basic test to assert gauge equality

        Test all gauges found in gauges.data.  Do full comparison of all
        times, levels, components of q.

        :Input:
         - *save* (bool) - If *True* will save the output from this test to 
           the file *regresion_data.txt*.  Default is *False*.
         - *regression_data_path* (path) - Path to the regression test data.
           Defaults to 'regression_data.txt'.
         - *tolerance* (float) - Tolerance used in comparison, defaults to
           *1e-14*.
        """
        from clawpack.visclaw import gaugetools
        from clawpack.visclaw.data import ClawPlotData

        # Get gauge data
        plotdata = ClawPlotData()
        plotdata.outdir = self.temp_path

        setgauges = gaugetools.read_setgauges(plotdata.outdir)
        gauge_numbers = setgauges.gauge_numbers
        
        # read in all gauge data and sort by gauge numbers so we
        # can properly compare.  Note gauges may be written to fort.gauge
        # in random order when OpenMP used.

        for gaugeno in gauge_numbers:
            g = plotdata.getgauge(gaugeno)
            m = len(g.level)
            number = numpy.array(m*[g.number])
            level = numpy.array(g.level)
            data_gauge = numpy.vstack((number,level,g.t,g.q)).T
            try:    
                data = numpy.vstack((data, data_gauge))
            except:
                data = data_gauge  # first time thru loop

        # save the gauge number sorted by gaugeno in case user wants to
        # compare if assertion fails.
        sorted_gauge_file = 'test_gauge.txt'
        sorted_gauge_path = os.path.join(self.temp_path, sorted_gauge_file)
        numpy.savetxt(sorted_gauge_path, data)

        # Get (and save) regression comparison data
        regression_data_file = os.path.join(self.test_path,
                                            regression_data_path)
        if save:
            numpy.savetxt(regression_data_file, data)
            print("Saved new regression data to %s" % regression_data_file)
        regression_data = numpy.loadtxt(regression_data_file)

        # if assertion fails, indicate to user what files to compare:
        output_dir = os.path.join(os.getcwd(),
                                     "%s_output" % self.__class__.__name__)
        sorted_gauge_path = os.path.join(output_dir, sorted_gauge_file)
        error_msg = "Full gauge match failed.  Compare these files: " + \
               "\n  %s\n  %s" % (regression_data_file, sorted_gauge_path) + \
               "\nColumns are gaugeno, level, t, q[0:num_eqn]"

        assert numpy.allclose(data, regression_data, tolerance), error_msg


    def tearDown(self):
        r"""Tear down test infrastructure.

        Closes *stdout* and *stderr*, removes the temporary directoy and if 
        *success* is *False*, copies the contents of the tempory directory to
        the current working directory.

        """
        self.stdout.close()
        self.stderr.close()

        if not self.success:
            output_dir = os.path.join(os.getcwd(),
                                         "%s_output" % self.__class__.__name__)

            # Remove directory if it exists already
            if os.path.exists(output_dir):
                shutil.rmtree(output_dir)

            # Copy out output files to local directory
            shutil.copytree(self.temp_path, output_dir)

        shutil.rmtree(self.temp_path)
        self.temp_path = None



"""
Regression tests for 2D advection on an annulus.
"""

from __future__ import absolute_import
import sys
import unittest

import clawpack.classic.test as test


class Advection2DAnnulusTest(test.ClassicRegressionTest):
    r"""Basic test for a 2D advection test case"""


    def runTest(self, save=False):

        # Write out data files
        self.load_rundata()
        self.write_rundata_objects()

        # Run code
        self.run_code()

        # Perform tests
        self.check_frame(save=save, frame_num=1, 
                         file_name='regression_data_test2.txt')
        self.check_frame(save=save, frame_num=2, 
                         file_name='regression_data_test3.txt')

        self.success = True



if __name__=="__main__":
    if len(sys.argv) > 1:
        if bool(sys.argv[1]):
            # Fake the setup and save out output
            test = Advection2DAnnulusTest()
            try:
                test.setUp()
                test.runTest(save=True)
            finally:
                test.tearDown()
            sys.exit(0)
    unittest.main()


#!/usr/bin/env python3
import re
import os
import sys


def get_name(settings=None):
    return "mpi4py"


def get_version(settings=None):
    topdir = os.path.dirname(os.path.abspath(__file__))
    source = os.path.join(topdir, "clawpack", "__init__.py")
    with open(source, encoding="utf-8") as f:
        m = re.search(r"__version__\s*=\s*'(.*)'", f.read())
    version = m.groups()[0]
    local_version = os.environ.get("CLAWPACK_LOCAL_VERSION")
    if local_version:
        version = "{version}+{local_version}".format(**vars())
    return version



def dynamic_metadata(field, settings=None):
    getter = globals().get("get_" + field)
    if getter:
        return getter(settings)
    return globals()[field.replace(".", "_")]


if __name__ == "__main__":
    print(dynamic_metadata(sys.argv[1]))




from __future__ import print_function
from pylab import *

def latlong(d,theta,phi,Rearth):
    """
    Take a point at distance d from the North pole with longitude theta
    and return longitude and latitude of point after rotating pole down 
    to latitude phi.
    Applying this function to a set of points on a circle of radius d 
    will result in points that are equi-distant (distance d on the earth) 
    from the point at latitude phi, longitude 0.  

    Used to construct a radially symmetric ocean on the earth and place 
    gauges, islands, etc.
    """

    # Convert phi to radians:
    theta = theta * pi/180.
    phi = phi * pi/180.

    alpha = pi/2. - d/Rearth   # latitude of original point
    # x,y,z coordinates of this point on the earth:
    x1 = cos(alpha)*cos(theta)
    y1 = cos(alpha)*sin(theta)
    z1 = sin(alpha)

    # rotate so centered at latitude phi:
    x2 = x1*sin(phi) + z1*cos(phi)
    y2 = y1
    z2 = -x1*cos(phi) + z1*sin(phi)

    # compute longitude xhat and latitude yhat:
    xhat = -arctan(y2/x2)
    yhat = arcsin(z2)

    # convert to degrees:
    xhat = xhat * 180./pi
    yhat = yhat * 180./pi

    return xhat,yhat

def plot_ocean_and_shelf(d1=1580e3, d2=1645e3, phi=40.):

    theta = linspace(0, 360., 200)
    d = d1*ones(theta.shape)
    xhat, yhat = latlong(d, theta, phi)

    clf()
    plot(xhat,yhat,'b')
    hold(True)

    d = d2*ones(theta.shape)
    xhat, yhat = latlong(d, theta, phi)

    plot(xhat,yhat,'r')
    legend(['Continental shelf', 'Shoreline'],loc='lower left')

    (xi1,yi1) = latlong(1600.e3,220.,40.)
    plot([xi1],[yi1],'ko')
    (xi2,yi2) = latlong(1600.e3,260.,40.)
    plot([xi2],[yi2],'ko')
    axis('scaled')
    xlim([-20, 20])
    ylim([15, 60])

#==========================================================
def gcdist(x1,y1,x2,y2,Rearth,units='degrees'):
    """
    Compute the great circle distance on the earth between points
    (x1,y1) and (x2,y2), where:
    x = longitude, y = latitude 

    Taken from Clawpack 4.x.
    """
    from numpy import pi,sin,cos,arccos,arcsin,sqrt
    if units=='degrees':
        # convert to radians:
        x1 = x1 * pi/180.
        y1 = y1 * pi/180.
        x2 = x2 * pi/180.
        y2 = y2 * pi/180.
    elif units != 'radians':
        raise Exception("unrecognized units")

    dx = x1 - x2
    dy = y1 - y2

    # angle subtended by two points, using Haversine formula:
    dsigma = 2. * arcsin(sqrt(sin(0.5*dy)**2 + cos(y1)*cos(y2)*sin(0.5*dx)**2))

    # alternative formula that may have more rounding error:
    #dsigma2 = arccos(sin(y1)*sin(y2)+ cos(y1)*cos(y2)*cos(dx))
    #print("max diff in dsigma: ", abs(dsigma-dsigma2).max())

    d = Rearth * dsigma
    return d





#!/usr/bin/env python

r"""
Module defines a class and routines for managing storm best-track type input and
testing reconstructed wind and pressure fields.  Additionally some support for
ensembles of storms from various providers is also included.

The primary class of interest in the module is the `Storm` class that
facilitates dealing with various best-track formats found around the world and
the expected GeoClaw storm format that is read into the FORTRAN code.  The basic
workflow in a `setrun.py` file would do the following:

1. Create a `Storm` object by reading in from a file::

    storm = clawpack.geoclaw.surge.storm.Storm("my_storm.txt", file_format='ATCF')

2. Write out the storm object created into the GeoClaw format::

    storm.write("my_geoclaw_storm.txt", file_format="geoclaw")

3. Specify the path to the GeoClaw formatted storm file, in this case
   "my_geoclaw_storm.txt".

:Formats Supported:
    - GeoClaw (fully)
    - ATCF (reading only)
    - HURDAT (reading only)
    - IBTrACS (reading only)
    - JMA (reading only)
    - IMD (planned)
    - tcvitals (reading only)
"""

import warnings
import sys
import os
import argparse
import datetime

import numpy as np
import pandas as pd

import clawpack.geoclaw.units as units

# =============================================================================
#  Common acronyms across formats

# ATCF basins with their expanded names
# see https://www.nrlmry.navy.mil/atcf_web/docs/database/new/abrdeck.html
ATCF_basins = {"AL": "Atlantic",
               "CP": "Central Pacific",
               "EP": "East Pacific",
               "IO": "North Indian Ocean",
               "SH": "Southern Hemisphere",
               "SL": "Southern Atlantic",
               "LS": "Southern Atlantic",
               "WP": "North West Pacific"}

# TCVitals basins with their expanded names
# see http://www.emc.ncep.noaa.gov/HWRF/tcvitals-draft.html
TCVitals_Basins = {"L": "North Atlantic",
                   "E": "North East Pacific",
                   "C": "North Central Pacific",
                   "W": "North West Pacific",
                   "B": "Bay of Bengal (North Indian Ocean)",
                   "A": "Arabian Sea (North Indian Ocean)",
                   "Q": "South Atlantic",
                   "P": "South Pacific",
                   "S": "South Indian Ocean"}

# Tropical Cyclone Designations
# see https://www.nrlmry.navy.mil/atcf_web/docs/database/new/abrdeck.html
TC_designations = {"DB": "disturbance",
                   "TD": "tropical depression",
                   "TS": "tropical storm",
                   "TY": "typhoon",
                   "ST": "super typhoon",
                   "TC": "tropical cyclone",
                   "HU": "hurricane",
                   "SD": "subtropical depression",
                   "SS": "subtropical storm",
                   "EX": "extratropical systems",
                   "IN": "inland",
                   "DS": "dissipating",
                   "LO": "low",
                   "WV": "tropical wave",
                   "ET": "extrapolated",
                   "XX": "unknown"}

# HURDAT special designations
# see http://www.aoml.noaa.gov/hrd/data_sub/newHURDAT.html
hurdat_special_entries = {"L": "landfall",
                          "W": "max wind",
                          "P": "min pressure",
                          "I": "max intensity",
                          "C": "closest approach",
                          "S": "status change",
                          "G": "genesis",
                          "T": "additional track point"}


# Warning for formats that have yet to have a default way to determine crticial
# radii from the input data
missing_data_warning_str = """*** Cannot yet automatically determine the
    maximum wind radius.  Will write out GeoClaw
    formats but note that these will not work
    when running GeoClaw currently without a custom
    `max_wind_radius_fill` function passed as argument
    to the `write` function."""

# Warning for not having any time points with both a max wind speed and central
# pressure observation
missing_necessary_data_warning_str = """No storm points in the input file
    had both a max wind speed and a central pressure observation."""


class NoDataError(ValueError):
    """Exception to raise when no valid data in input file"""
    pass


# =============================================================================
#  Basic storm class
class Storm(object):
    r"""
    Storm data object

    This object contains a time series of time data that describe a particular
    storm.  This includes the attributes below and the ability to read from
    multiple sources for data such as the U.S. National Hurricane Center (NHC),
    the Japanese Meterological Agency (JMA), and the Indian Meteorlogical
    Department (IMD).  This class can then write out in any of these formats,
    construct the wind and pressure fields using a supported parameterized
    model, or output the GeoClaw supported storm format used for running storm
    surge simulations.

    *TODO:*  Add description of unit handling

    :Attributes:
     - *t* (list(float) or list(datetime.datetiem)) Contains the time at which
       each entry of the other arrays are at.  These are expected to
       be *datetime* objects. Note that when written some formats require
       a *time_offset* to be set.
     - *eye_location* (ndarray(:, :)) location of the eye of the storm. Default
       units are in signed decimal longitude and latitude.
     - *max_wind_speed* (ndarray(:)) Maximum wind speed.  Default units are
       meters/second.
     - *max_wind_radius* (ndarray(:)) Radius at which the maximum wind speed
       occurs.  Default units are meters.
     - *central_pressure* (ndarray(:)) Central pressure of storm.  Default units
       are Pascals.
     - *storm_radius* (ndarray(:)) Radius of storm, often defined as the last
       closed iso-bar of pressure.  Default units are meters.
     - *time_offset* (datetime.datetime) A date time that as an offset for the
       simulation time.  This will default to the beginning of the first of
       the year that the first time point is found in.
     - *wind_speeds* (ndarray(:, :)) Wind speeds defined in every record, such
       as 34kt, 50kt, 64kt, etc and their radii. Default units are
       meters/second and meters.

    :Initialization:
     1. Read in existing file at *path*.
     2. Construct an empty storm and supply the fields needed.  Note that these
        fields must be converted to the appropriate units.

    :Input:
     - *path* (string) Path to file to be read in if requested.
     - *file_format* (string) Format of file at path.  Default is "hurdat"
     - *kwargs* (dict) Other key-word arguments are passed to the appropriate
       read routine.
    """

    # Define supported formats and models - keys are function name related and
    # values are the proper name and a citation or URL documenting the format
    _supported_formats = {"geoclaw": ["GeoClaw", "http://www.clawpack.org/storms"],
                          "atcf": ["ATCF", "http://www.nrlmry.navy.mil/atcf_web/docs/database/new/database.html"],
                          "hurdat": ["HURDAT", "http://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html"],
                          "ibtracs": ["IBTrACS", "https://www.ncdc.noaa.gov/ibtracs/index.php?name=ib-v4-access"],
                          "jma": ["JMA", "http://www.jma.go.jp/jma/jma-eng/jma-center/rsmc-hp-pub-eg/Besttracks/e_format_bst.html"],
                          "imd": ["IMD", "http://www.rsmcnewdelhi.imd.gov.in/index.php"],
                          "tcvitals": ["TC-Vitals", "http://www.emc.ncep.noaa.gov/mmb/data_processing/tcvitals_description.htm"]}

    def __init__(self, path=None, file_format="ATCF", **kwargs):
        r"""Storm Initiatlization Routine

        See :class:`Storm` for more info.
        """

        self.t = None
        self.time_offset = None
        self.eye_location = None
        self.max_wind_speed = None
        self.max_wind_radius = None
        self.central_pressure = None
        self.storm_radius = None
        self.wind_speeds = None

        # Storm descriptions - not all formats provide these
        self.name = None                    # Possibly a list of a storm's names
        self.basin = None                   # Basin containing storm
        self.ID = None                      # ID code - depends on format
        self.classification = None          # Classification of storm (e.g. HU)
        self.event = None                   # Event (e.g. landfall) - HURDAT

        if path is not None:
            self.read(path, file_format=file_format, **kwargs)

    # ==========================================================================
    #  Basic object support
    def __str__(self):
        r""""""
        output = f"Name: {self.name}\n"
        if isinstance(self.t[0], datetime.datetime):
            output += f"Dates: {self.t[0].isoformat()}"
            output += f" - {self.t[-1].isoformat()}"
        else:
            output += f"Dates: {self.t[0]} - {self.t[-1]}"
        return output

    def __repr__(self):
        return '<{}.{} "{}" at {}>'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.__dict__.get('name', 'name not given'),
            hex(id(self)))

    # ==========================================================================
    # Read Routines
    def read(self, path=None, file_format="atcf", **kwargs):
        r"""Read in storm data from *path* with format *file_format*

        :Input:
         - *path* (string) Path to data file.
         - *file_format* (string) Format of the data file.  See list of
           supported formats for a list of valid strings.  Defaults to
           "hurdat".
         - *kwargs* (dict) Keyword dictionary for additional arguments that can
           be passed down to the appropriate read functions.  Please refer to
           the specific routine for a list of valid options.

        :Raises:
         - *ValueError* If the *file_format* requested does not match any of
           the available supported formats a *ValueError* is raised.
        """

        # If a path is not provided then we can try and find the relevant
        # database and download it
        if path is None:
            data_str = ("Currently automatic download of storm databases is ",
                        "not implemented.  Please refer to the URLs below for",
                        "references as to where you can download storm data",
                        "files:",
                        " - ATCF - http://ftp.nhc.noaa.gov/atcf/archive/",
                        " - HURDAT - http://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html",
                        " - IBTrACS - https://www.ncdc.noaa.gov/ibtracs/index.php?name=ib-v4-access",
                        " - JMA - http://www.jma.go.jp/jma/jma-eng/jma-center/rsmc-hp-pub-eg/besttrack.html",
                        " - IMD - http://www.rsmcnewdelhi.imd.gov.in/index.php",
                        " - TCVITALS - http://www.emc.ncep.noaa.gov/mmb/data_processing/tcvitals_description.htm")
            raise NotImplementedError("\n".join(data_str))

        if file_format.lower() not in self._supported_formats.keys():
            raise ValueError("File format %s not available." % file_format)

        getattr(self, 'read_%s' % file_format.lower())(path, **kwargs)

    def read_geoclaw(self, path, verbose=False):
        r"""Read in a GeoClaw formatted storm file

        GeoClaw storm files are read in by the Fortran code and are not meant
        to be human readable.

        :Input:
         - *path* (string) Path to the file to be read.
         - *verbose* (bool) Output more info regarding reading.
        """

        # Attempt to get name from file if is follows the convention name.storm
        if ".storm" in os.path.splitext(path):
            self.name = os.path.splitext(os.path.basename(path))[0]

        # Read header
        with open(path, 'r') as data_file:
            num_casts = int(data_file.readline())
            time = data_file.readline()[:19]
            try:
                self.time_offset = datetime.datetime.strptime(
                    time, '%Y-%m-%dT%H:%M:%S')
            except ValueError:
                self.time_offset = float(time)
        # Read rest of data
        data = np.loadtxt(path, skiprows=3)
        num_forecasts = data.shape[0]
        self.eye_location = np.empty((num_forecasts, 2))
        assert(num_casts == num_forecasts)
        if isinstance(self.time_offset, datetime.datetime):
            self.t = np.array([self.time_offset
                               + datetime.timedelta(seconds=data[i, 0])
                               for i in range(num_forecasts)])
        else:
            self.t = data[:, 0]
        self.eye_location[:, 0] = data[:, 1]
        self.eye_location[:, 1] = data[:, 2]
        self.max_wind_speed = data[:, 3]
        self.max_wind_radius = data[:, 4]
        self.central_pressure = data[:, 5]
        self.storm_radius = data[:, 6]

    def read_atcf(self, path, verbose=False):
        r"""Read in a ATCF formatted storm file

        ATCF format has storm stored individually so there is no support for
        multiple storms in a particular file.

        :Input:
         - *path* (string) Path to the file to be read.
         - *verbose* (bool) Output more info regarding reading.
        """

        # See here for the ATCF format documentation:
        #   https://www.nrlmry.navy.mil/atcf_web/docs/database/new/abdeck.txt

        # Slightly more robust converter for ATCF data fields that can be
        # missing
        def num_converter(x):
            if isinstance(x, str):
                if len(x.strip()) == 0:
                    # Only whitespace
                    return np.nan
                else:
                    # Assume this is still a number
                    return float(x)
            elif x is None:
                return np.nan
            return float(x)

        df = pd.read_csv(path, engine="python", sep=",+", names=[
            "BASIN", "CY", "YYYYMMDDHH", "TECHNUM", "TECH", "TAU",
            "LAT", "LON", "VMAX", "MSLP", "TY",
            "RAD", "WINDCODE", "RAD1", "RAD2", "RAD3", "RAD4",
            "POUTER", "ROUTER", "RMW", "GUSTS", "EYE", "SUBREGION",
            "MAXSEAS", "INITIALS", "DIR", "SPEED", "STORMNAME", "DEPTH",
            "SEAS", "SEASCODE", "SEAS1", "SEAS2", "SEAS3", "SEAS4",
            "USERDEFINE1", "userdata1",
            "USERDEFINE2", "userdata2",
            "USERDEFINE3", "userdata3",
            "USERDEFINE4", "userdata4",
            "USERDEFINE5", "userdata5",
        ],
            converters={
                "YYYYMMDDHH": lambda d: datetime.datetime(
                    int(d[1:5]), int(d[5:7]), int(d[7:9]), int(d[9:11])),
                "TAU": lambda d: datetime.timedelta(hours=int(d)),
                "LAT": lambda d: (-.1 if d[-1] == "S" else .1) * int(d.strip("NS ")),
                "LON": lambda d: (-.1 if d[-1] == "W" else .1) * int(d.strip("WE ")),
                "RAD": num_converter,
                "RAD": num_converter,
                "RAD1": num_converter,
                "RAD2": num_converter,
                "RAD3": num_converter,
                "RAD4": num_converter,
                "ROUTER": num_converter,
                "RMW": num_converter,
                "STORMNAME": lambda d: (d.strip() if isinstance(d, str) else d)
        },
            dtype={
                "BASIN": str,
                "CY": int,
                "VMAX": float,
                "MSLP": float,
                "TY": str
        })

        # Grab data regarding basin and cyclone number from first row
        self.basin = ATCF_basins[df["BASIN"][0]]
        self.ID = df["CY"][0]

        # Keep around the name as an array
        self.name = df["STORMNAME"].to_numpy()

        # Take forecast period TAU into consideration
        df['DATE'] = df["YYYYMMDDHH"] + df["TAU"]
        df = df[["DATE", "TAU", "TY", "LAT", "LON", "VMAX", "MSLP",
                 "ROUTER", "RMW", "RAD", "RAD1", "RAD2", "RAD3", "RAD4", ]]
        df = df.sort_values(by=["DATE", "TAU"]).reset_index(drop=True)

        # For each DATE, choose best (smallest TAU) available data
        for c in ["LAT", "LON", "VMAX", "MSLP", "ROUTER", "RMW",
                  "RAD", "RAD1", "RAD2", "RAD3", "RAD4"]:
            df[c] = df[c].where(df[c] != 0, np.nan)  # value 0 means NaN
            df[c] = df.groupby("DATE")[c].bfill()
        df = df.groupby("DATE").first()

        # Wind profile (occasionally missing for older ATCF storms)
        # Wind speeds and their radii
        df["RAD_MEAN"] = df[["RAD1", "RAD2", "RAD3", "RAD4"]].mean(
            axis=1, skipna=True)
        df = df.drop(["TAU", "RAD1", "RAD2", "RAD3", "RAD4"], axis=1)
        df = df.dropna(how="any", subset=["LAT", "LON"])

        # Create time
        self.t = list(df.index.to_pydatetime())

        # Classification, note that this is not the category of the storm
        self.classification = df["TY"].to_numpy()

        # Eye location - longitude/latitude order
        self.eye_location = df[["LON", "LAT"]].to_numpy()

        # Convert to correct units:
        #  max_wind_speed - Convert knots to m/s - 0.51444444
        #  max_wind_radius  - convert from nm to m - 1.8520000031807990 * 1000.0
        #  central_pressure - convert from mbar to Pa - 100.0
        #  Radius of last isobar contour - convert from nm to m - 1.852000003180799d0 * 1000.0
        self.max_wind_speed = units.convert(
            df["VMAX"].to_numpy(), 'knots', 'm/s')
        self.central_pressure = units.convert(
            df["MSLP"].to_numpy(), 'mbar', 'Pa')
        self.max_wind_radius = units.convert(df["RMW"].to_numpy(), 'nmi', 'm')
        self.storm_radius = units.convert(df["ROUTER"].to_numpy(), 'nmi', 'm')
        self.wind_speeds = df[["RAD", "RAD_MEAN"]].to_numpy()
        self.wind_speeds[:, 0] = units.convert(
            self.wind_speeds[:, 0], 'knots', 'm/s')
        self.wind_speeds[:, 1] = units.convert(
            self.wind_speeds[:, 1], 'nmi', 'm')

    def read_hurdat(self, path, verbose=False):
        r"""Read in HURDAT formatted storm file

        This is the current version of HURDAT data available (HURDAT 2).  Note
        that this assumes there is only one storm in the file (includes the
        header information though).  Future features will be added that will allow for
        a file to be read with multiple storms defined.

        For more details on the HURDAT format and getting data see

        http://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html

        :Input:
         - *path* (string) Path to the file to be read.
         - *verbose* (bool) Output more info regarding reading.

        :Raises:
         - *ValueError* If the method cannot find the name/year matching the
           storm or they are not provided when *single_storm == False* then a
           value error is risen.
        """

        with open(path, 'r') as hurdat_file:
            # Extract header
            data = [value.strip() for value in
                    hurdat_file.readline().split(',')]
            self.basin = data[0][:2]
            self.name = data[1]
            self.ID = data[2]

            # Store rest of data
            data_block = hurdat_file.readlines()

        num_lines = len(data_block)

        # Parse data block
        self.t = []
        self.event = np.empty(num_lines, dtype=str)
        self.classification = np.empty(num_lines, dtype=str)
        self.eye_location = np.empty((num_lines, 2))
        self.max_wind_speed = np.empty(num_lines)
        self.central_pressure = np.empty(num_lines)
        self.max_wind_radius = np.empty(num_lines)
        self.storm_radius = np.empty(num_lines)

        for (i, line) in enumerate(data_block):
            if len(line) == 0:
                break
            data = [value.strip() for value in line.split(",")]

            # Create time
            self.t.append(datetime.datetime(int(data[0][:4]),
                                            int(data[0][4:6]),
                                            int(data[0][6:8]),
                                            int(data[1][:2]),
                                            int(data[1][2:])))

            # If an event is occuring record it.  If landfall then use as an
            # offset.   Note that if there are multiple landfalls the last one
            # is used as the offset
            if len(data[2].strip()) > 0:
                self.event[i] = data[2].strip()
                if self.event[i].upper() == "L":
                    self.time_offset = self.t[i]

            # Classification, note that this is not the category of the storm
            self.classification[i] = data[3]

            # Parse eye location
            if data[4][-1] == "N":
                self.eye_location[i, 1] = float(data[4][0:-1])
            else:
                self.eye_location[i, 1] = -float(data[4][0:-1])
            if data[5][-1] == "E":
                self.eye_location[i, 0] = float(data[5][0:-1])
            else:
                self.eye_location[i, 0] = -float(data[5][0:-1])

            # Intensity information - radii are not included directly in this
            # format and instead radii of winds above a threshold are included
            self.max_wind_speed[i] = units.convert(
                float(data[6]), 'knots', 'm/s')
            self.central_pressure[i] = units.convert(
                float(data[7]), 'mbar', 'Pa')
            warnings.warn(missing_data_warning_str)
            self.max_wind_radius[i] = -1
            self.storm_radius[i] = -1

    def read_ibtracs(self, path, sid=None, storm_name=None, year=None, start_date=None,
                     agency_pref=['wmo',
                                  'usa',
                                  'tokyo',
                                  'newdelhi',
                                  'reunion',
                                  'bom',
                                  'nadi',
                                  'wellington',
                                  'cma',
                                  'hko',
                                  'ds824',
                                  'td9636',
                                  'td9635',
                                  'neumann',
                                  'mlc']):
        r"""Read in IBTrACS formatted storm file

        This reads in the netcdf-formatted IBTrACS v4 data. You must either pass
        the *sid* of the storm (a unique identifier supplied by IBTrACS) OR
        *storm_name* and *year*. The latter will not be unique for unnamed storms,
        so you may optionally pass *start_date* as well. The `wmo_\*` variable is
        used when non-missing, with missing values filled in by the corresponding
        variable of the agency specified in `wmo_agency` and/or `usa_agency`. If
        still missing, the other agencies are checked in order of *agency_pref* to
        see if any more non-missing values are available.

        :Input:
         - *path* (string) Path to the file to be read.
         - *sid* (string, optional) IBTrACS-supplied unique track identifier.
             Either *sid* OR *storm_name* and *year* must not be None.
         - *storm_name* (string, optional) name of storm of interest
             (NAME field in IBTrACS). Either *sid* OR *storm_name* and
             *year* must not be None.
         - *year* (int, optional) year of storm of interest.
             Either *sid* OR *storm_name* and *year* must not be None.
         - *start_date* (:py:class:`datetime.datetime`, optional) If storm is not
             named, will find closest unnamed storm to this start date. Only
             used for unnamed storms when specifying *storm_name* and *year*
             does not uniquely identify storm.
         - *agency_pref* (list, optional) Preference order to use if `wmo_\*` variable
             is missing and `wmo_agency` and `usa_agency` are also missing.

        :Raises:
         - *ValueError* If the method cannot find the matching storm then a
             value error is risen.
        """

        # imports that you don't need for other read functions
        try:
            import xarray as xr
        except ImportError as e:
            print("IBTrACS currently requires xarray to work.")
            raise e

        # only allow one method for specifying storms
        if (sid is not None) and ((storm_name is not None) or (year is not None)):
            raise ValueError(
                'Cannot specify both *sid* and *storm_name* or *year*.')

        with xr.open_dataset(path) as ds:

            # match on sid
            if sid is not None:
                match = ds.sid == sid.encode()
            # or match on storm_name and year
            else:
                storm_name = storm_name.upper()
                # in case storm is unnamed
                if storm_name.upper() in ['UNNAMED', 'NO-NAME']:
                    storm_name = 'NOT_NAMED'
                storm_match = (ds.name == storm_name.encode())
                year_match = (ds.time.dt.year == year).any(dim='date_time')
                match = storm_match & year_match
            ds = ds.sel(storm=match).squeeze()

            # occurs if we have 0 or >1 matching storms
            if 'storm' in ds.dims.keys():
                if ds.storm.shape[0] == 0:
                    raise ValueError('Storm/year not found in provided file')
                else:
                    # see if a date was provided for multiple unnamed storms
                    assert start_date is not None, ValueError(
                        'Multiple storms identified and no start_date specified.')

                    start_times = ds.time.isel(date_time=0)
                    start_date = np.datetime64(start_date)

                    # find storm with start date closest to provided
                    storm_ix = abs(start_times - start_date).argmin()
                    ds = ds.isel(storm=storm_ix).squeeze()
                    assert 'storm' not in ds.dims.keys()

            # cut down dataset to only non-null times
            valid_t = ds.time.notnull()
            if valid_t.sum() == 0:
                raise ValueError('No valid wind speeds found for this storm.')
            ds = ds.sel(date_time=valid_t)

            # list of the agencies that correspond to 'usa_*' variables
            usa_agencies = [b'atcf', b'hurdat_atl', b'hurdat_epa', b'jtwc_ep',
                            b'nhc_working_bt', b'tcvightals', b'tcvitals']

            # Create mapping from wmo_ or usa_agency
            # to the appropriate variable
            agency_map = {b'': agency_pref.index('wmo')}
            # account for multiple usa agencies
            for a in usa_agencies:
                agency_map[a] = agency_pref.index('usa')
            # map all other agencies to themselves
            for i in [a for a in agency_pref if a not in ['wmo', 'usa']]:
                agency_map[i.encode('utf-8')] = agency_pref.index(i)

            # fill in usa as provider if usa_agency is
            # non-null when wmo_agency is null
            provider = ds.wmo_agency.where(ds.wmo_agency != b'', ds.usa_agency)

            # get index into from agency that is wmo_provider
            def map_val_to_ix(a):
                def func(x): return agency_map[x]
                return xr.apply_ufunc(func, a, vectorize=True)
            pref_agency_ix = map_val_to_ix(provider)

            # GET MAX WIND SPEED and PRES
            pref_vals = {}
            for v in ['wind', 'pres']:
                all_vals = ds[['{}_{}'.format(i, v) for i in agency_pref]].to_array(
                    dim='agency')

                # get wmo value
                val_pref = ds['wmo_'+v]

                # fill this value in as a second-best
                pref_2 = all_vals.isel(agency=pref_agency_ix)
                val_pref = val_pref.fillna(pref_2)

                # now use the agency_pref order to fill in
                # any remaining values as third best
                best_ix = all_vals.notnull().argmax(dim='agency')
                pref_3 = all_vals.isel(agency=best_ix)
                val_pref = val_pref.fillna(pref_3)

                # add to dict
                pref_vals[v] = val_pref

            # THESE CANNOT BE MISSING SO DROP
            # IF EITHER MISSING
            valid = pref_vals['wind'].notnull() & pref_vals['pres'].notnull()
            if not valid.any():
                raise NoDataError(missing_necessary_data_warning_str)
            ds = ds.sel(date_time=valid)
            for i in ['wind', 'pres']:
                pref_vals[i] = pref_vals[i].sel(date_time=valid)

            # GET RMW and ROCI
            # (these can be missing)
            for r in ['rmw', 'roci']:
                order = ['{}_{}'.format(i, r) for i in agency_pref if
                         '{}_{}'.format(i, r) in ds.data_vars.keys()]
                vals = ds[order].to_array(dim='agency')
                best_ix = vals.notnull().argmax(dim='agency')
                val_pref = vals.isel(agency=best_ix)
                pref_vals[r] = val_pref

            # CONVERT TO GEOCLAW FORMAT

            # assign basin to be the basin where track originates
            # in case track moves across basins
            self.basin = ds.basin.values[0].astype(str)
            self.name = ds.name.astype(str).item()
            self.ID = ds.sid.astype(str).item()

            # convert datetime64 to datetime.datetime
            self.t = []
            for d in ds.time:
                t = d.dt
                self.t.append(datetime.datetime(t.year, t.month,
                                                t.day, t.hour, t.minute, t.second))

            # events
            self.event = ds.usa_record.values.astype(str)

            # time offset
            if (self.event == 'L').any():
                # if landfall, use last landfall
                self.time_offset = np.array(self.t)[self.event == 'L'][-1]
            else:
                # if no landfall, use last time of storm
                self.time_offset = self.t[-1]

            # Classification, note that this is not the category of the storm
            self.classification = ds.usa_status.values
            self.eye_location = np.array([ds.lon, ds.lat]).T

            # Intensity information - for now, including only common, basic intensity
            # info.
            # TODO: add more detailed info for storms that have it
            self.max_wind_speed = units.convert(
                pref_vals['wind'], 'knots', 'm/s').where(pref_vals['wind'].notnull(), -1).values
            self.central_pressure = units.convert(pref_vals['pres'], 'mbar', 'Pa').where(
                pref_vals['pres'].notnull(), -1).values
            self.max_wind_radius = units.convert(pref_vals['rmw'], 'nmi', 'm').where(
                pref_vals['rmw'].notnull(), -1).values
            self.storm_radius = units.convert(pref_vals['roci'], 'nmi', 'm').where(
                pref_vals['roci'].notnull(), -1).values

            # warn if you have missing vals for RMW or ROCI
            if (self.max_wind_radius.max()) == -1 or (self.storm_radius.max() == -1):
                warnings.warn(missing_data_warning_str)

    def read_jma(self, path, verbose=False):
        r"""Read in JMA formatted storm file

        Note that only files that contain one storm are currently supported.

        For more details on the JMA format and getting data see

        http://www.jma.go.jp/jma/jma-eng/jma-center/rsmc-hp-pub-eg/Besttracks/e_format_bst.html

        :Input:
         - *path* (string) Path to the file to be read.
         - *verbose* (bool) Output more info regarding reading.

        :Raises:
         - *ValueError* If the method cannot find the name/year matching the
           storm or they are not provided when *single_storm == False* then a
           value error is risen.
        """

        data_block = []
        with open(path, 'r') as JMA_file:
            # Extract header
            data = JMA_file.readline()
            self.ID = data[6:10]
            num_lines = int(data[12:14])
            self.name = data[30:51].strip()

            data_block = JMA_file.readlines()
        assert(num_lines == len(data_block))

        # Parse data block
        self.t = []
        self.event = np.empty(num_lines, dtype=str)
        self.classification = np.empty(num_lines, dtype=str)
        self.eye_location = np.empty((num_lines, 2))
        self.max_wind_speed = np.empty(num_lines)
        self.central_pressure = np.empty(num_lines)
        self.max_wind_radius = np.empty(num_lines)
        self.storm_radius = np.empty(num_lines)
        for (i, line) in enumerate(data_block):
            if len(line) == 0:
                break
            data = [value.strip() for value in line.split()]

            # Create time
            self.t.append(datetime.datetime(int(data[0][:2]),
                                            int(data[0][2:4]),
                                            int(data[0][4:6]),
                                            int(data[0][6:])))

            # Classification, note that this is not the category of the storm
            self.classification[i] = int(data[1])

            # Parse eye location - Always N latitude and E longitude
            self.eye_location[i, 0] = float(data[4]) / 10.0
            self.eye_location[i, 1] = float(data[3]) / 10.0

            # Intensity information - current the radii are not directly given
            # Available data includes max/min of radius of winds of 50 and
            # 30 kts instead
            self.central_pressure[i] = units.convert(
                float(data[5]), 'hPa', 'Pa')
            self.max_wind_speed[i] = units.convert(
                float(data[6]), 'knots', 'm/s')
            warnings.warn(missing_data_warning_str)
            self.max_wind_radius[i] = -1
            self.storm_radius[i] = -1

    def read_imd(self, path, verbose=False):
        r"""Extract relevant hurricane data from IMD file
            and update storm fields with proper values.

        :Input:
         - *path* (string) Path to the file to be read.

        Return ValueError if format incorrect or if file not IMD.
        """
        raise NotImplementedError(("Reading in IMD files is not ",
                                   "implemented yet but is planned for a ",
                                   "future release."))

    def read_tcvitals(self, path, verbose=False):
        r"""Extract relevant hurricane data from TCVITALS file
            and update storm fields with proper values.

        :Input:
         - *path* (string) Path to the file to be read.
         - *verbose* (bool) Output more info regarding reading.

        """

        # read in TCVitals_file
        data_block = []
        with open(path, 'r') as TCVitals_file:
            data = TCVitals_file.readlines()
            for line in data:
                line = line.split()
                line = [value.strip() for value in line]
                data_block.append(line)
        num_lines = len(data_block)

        # Parse data block - convert to correct units
        # Conversions:
        #  max_wind_radius  - convert from km to m - 1000.0
        #  Central_pressure - convert from mbar to Pa - 100.0
        #  Radius of last isobar contour - convert from km to m - 1000.0
        self.t = []
        self.classification = np.empty(num_lines, dtype=str)
        self.eye_location = np.empty((num_lines, 2))
        self.max_wind_speed = np.empty(num_lines)
        self.central_pressure = np.empty(num_lines)
        self.max_wind_radius = np.empty(num_lines)
        self.storm_radius = np.empty(num_lines)

        for (i, data) in enumerate(data_block):
            # End at an empty lines - skips lines at the bottom of a file
            if len(data) == 0:
                break

            # Grab data regarding basin and cyclone number if we are starting
            if i == 0:
                self.basin = TCVitals_Basins[data[1][2:]]
                self.ID = int(data[1][:2])

            # Create time
            self.t.append(datetime.datetime(int(data[3][0:4]),
                                            int(data[3][4:6]),
                                            int(data[3][6:]),
                                            int(data[4][:2])))

            # Parse eye location - longitude/latitude order
            if data[5][-1] == 'N':
                self.eye_location[i, 1] = float(data[5][0:-1])/10.0
            else:
                self.eye_location[i, 1] = -float(data[5][0:-1])/10.0
            if data[6][-1] == "E":
                self.eye_location[i, 0] = float(data[6][0:-1])/10.0
            else:
                self.eye_location[i, 0] = -float(data[6][0:-1])/10.0

            # Intensity Information
            self.max_wind_speed[i] = float(data[12])
            self.central_pressure[i] = units.convert(
                float(data[9]), 'mbar', 'Pa')
            self.max_wind_radius[i] = units.convert(float(data[13]), 'km', 'm')
            self.storm_radius[i] = units.convert(float(data[11]), 'km', 'm')

    # =========================================================================
    # Write Routines

    def write(self, path, file_format="geoclaw", **kwargs):
        r"""Write out the storm data to *path* in format *file_format*

        :Input:
         - *path* (string) Path to data file.
         - *file_format* (string) Format of the data file.  See list of
           supported formats for a list of valid strings.  Defaults to
           "geoclaw".
         - *kwargs* (dict) Keyword dictionary for additional arguments that can
           be passed down to the appropriate write functions.  Please refer to
           the specific routine for a list of valid options.

        :Raises:
         - *ValueError* If the *file_format* requested does not match any of
           the available supported formats a *ValueError* is raised.
        """

        if file_format.lower() not in self._supported_formats.keys():
            raise ValueError("File format %s not available." % file_format)

        getattr(self, 'write_%s' % file_format.lower())(path, **kwargs)

    def write_geoclaw(self, path, force=False, skip=True, verbose=False,
                      fill_dict={}, **kwargs):
        r"""Write out a GeoClaw formatted storm file

        GeoClaw storm files are read in by the GeoClaw Fortran code.

        :Input:
         - *path* (string) Path to the file to be written.
         - *skip* (bool) Skip a time if NaNs are found and are not replaced.  
            Default is `True`.
         - *force* (bool) Force output of storm even if there is missing data.
            Default is `False`.
         - *verbose* (bool) Print out additional information when writing.
            Default is `False`.
         - *fill_dict* (dict) Dictionary of functions to use to fill in missing
            data represented by NaNs.  The keys are the field to be filled and
            the function signature should be `my_func(t, storm)` where t is the
            time of the forecast and `storm` is the storm object.  If the
            field remains a NaN or a function is not provided these lines will
            be assumed redundant and will be ommitted.  Note that the older 
            keyword arguments are put in this dictionary.  Currently the one
            default function is for `storm_radius`, which sets the value to 
            500 km.
        """

        # If a filling function is not provided we will provide some defaults
        fill_dict.update({"storm_radius": lambda t, storm: 500e3})
        # Handle older interface that had specific fill functions
        if "max_wind_radius_fill" in kwargs.keys():
            fill_dict.update(
                {"max_wind_radius": kwargs['max_wind_radius_fill']})
        if "storm_radius_fill" in kwargs.keys():
            fill_dict.update({"storm_radius": kwargs['storm_radius_fill']})

        # Loop through each line of data and if the line is valid, perform the
        # necessary work to write it out.  Otherwise either raise an exception
        # or skip it
        num_casts = 0
        data = []
        for n in range(len(self.t)):
            if self.t[n] == self.t[n - 1]:
                # Skip this time
                continue

            # Check each value we need for this time to make sure it is valid
            valid = True
            for name in ["max_wind_speed", "central_pressure",
                         "max_wind_radius", "storm_radius"]:
                if np.isnan(getattr(self, name)[n]):
                    if name in fill_dict.keys():
                        # Fill value with function provided
                        getattr(self, name)[n] = fill_dict[name](
                            self.t[n], self)
                    elif skip:
                        # Skip this line
                        valid = False
                        if verbose:
                            # Just warn that a NaN was found but continue
                            msg = ("*** WARNING:  The value {} at {} is a " +
                                   "NaN. Skipping this line.")
                            warnings.warn(msg.format(name, self.t[n]))
                    elif not force:
                        # If we are not asked to force to write raise an
                        # exception given the NaN
                        msg = ("The value {} at {} is a NaN and the storm " +
                               "will not be written in GeoClaw format.  If " +
                               "you want to fill in the value provide a " +
                               "function or set `force=True`.")
                        raise ValueError(msg.format(name, self.t[n]))
            if not valid:
                continue

            # Succeeded, add this time to the output
            num_casts += 1
            data.append(np.empty(7))

            # If we do not have a time offset use the first valid row as the
            # offset time
            if self.time_offset is None:
                self.time_offset = self.t[n]

            # Time
            if not isinstance(self.time_offset, float):
                data[-1][0] = (self.t[n] - self.time_offset).total_seconds()
            else:
                data[-1][0] = self.t[n] - self.time_offset
            # Eye-location
            data[-1][1:3] = self.eye_location[n, :]
            # Max wind speed
            data[-1][3] = self.max_wind_speed[n]
            # Max wind radius
            data[-1][4] = self.max_wind_radius[n]
            # Central pressure
            data[-1][5] = self.central_pressure[n]
            # Outer storm radius
            data[-1][6] = self.storm_radius[n]

        # Write out file
        format_string = ("{:19,.8e} " * 7)[:-1] + "\n"
        try:
            with open(path, "w") as data_file:
                # Write header
                data_file.write(f"{num_casts}\n")
                if isinstance(self.time_offset, datetime.datetime):
                    data_file.write(f"{self.time_offset.isoformat()}\n\n")
                else:
                    data_file.write(f"{str(self.time_offset)}\n\n")

                # Write data lines
                for line in data:
                    data_file.write(format_string.format(*line))

        except Exception as e:
            # If an exception occurs clean up a partially generated file
            if os.path.exists(path):
                os.remove(path)
            raise e

    def write_atcf(self, path, verbose=False):
        r"""Write out a ATCF formatted storm file

        :Input:
         - *path* (string) Path to the file to be written.
         - *verbose* (bool) Print out additional information when writing.
        """
        raise NotImplementedError(("Writing out ATCF files is not implemented ",
                                   "yet but is planned for a future release."))
        try:
            with open(path, 'w') as data_file:
                for n in range(len(self.t)):
                    data_file.write("".join((", " * 2,
                                             "%s" % seconds2date(self.t[n]),
                                             ", " * 4,
                                             "%s" % (int(self.eye_location[n, 0] *
                                                         10.0)),
                                             ", ",
                                             "%s" % (int(self.eye_location[n, 1] *
                                                         10.0)),
                                             ", ",
                                             "%s" % self.max_wind_speed[n],
                                             ", ",
                                             "%s" % self.central_pressure[n],
                                             ", ",
                                             ", " * 8,
                                             "%s" % self.storm_radius[n],
                                             ", ",
                                             "%s" % self.max_wind_radius[n],
                                             ", " * 10,
                                             "\n")))
        except Exception as e:
            # Remove possiblly partially generated file if not successful
            if os.path.exists(path):
                os.remove(path)
            raise e

    def write_hurdat(self, path, verbose=False):
        r"""Write out a HURDAT formatted storm file

        :Input:
         - *path* (string) Path to the file to be written.
         - *verbose* (bool) Print out additional information when writing.
        """
        raise NotImplementedError(("Writing out hurdat files is not ",
                                   "implemented yet but is planned for a ",
                                   "future release."))
        try:
            with open(path, 'w') as data_file:
                data_file.write('%s %s %s' % ("Date", "Hurricane Name",
                                              "Indicator"))
                for n in range(self.t.shape[0]):

                    latitude = float(self.eye_location[n, 0])
                    longitude = float(self.eye_location[n, 1])

                    # Convert latitude to proper Hurdat format e.g 12.0N
                    if latitude > 0:
                        latitude = str(np.abs(latitude)) + 'N'
                    else:
                        latitude = str(np.abs(latitude)) + 'S'

                    # Convert longitude to proper Hurdat format e.g 12.0W
                    if longitude > 0:
                        longitude = str(np.abs(longitude)) + 'E'
                    else:
                        longitude = str(np.abs(longitude)) + 'W'

                    data_file.write("".join(("%s" % self.seconds2date(
                        self.t[n])[0:-2],
                        "%s00" % self.seconds2date(
                        self.t[n])[-2:],
                        ", " * 3,
                        "%s" % (latitude),
                        ", ",
                        "%s" % (longitude),
                        ", ",
                        "%s" % self.max_wind_speed[n],
                        ", ",
                        "%s" % self.central_pressure[n],
                        ", ",
                        "%s" % self.storm_radius[n],
                        ", ",
                        "%s" % self.max_wind_radius[n],
                        ", " * 10,
                        "\n")))
        except Exception as e:
            # Remove possiblly partially generated file if not successful
            if os.path.exists(path):
                os.remove(path)
            raise e

    def write_jma(self, path, verbose=False):
        r"""Write out a JMA formatted storm file

        :Input:
         - *path* (string) Path to the file to be written.
         - *verbose* (bool) Print out additional information when writing.
        """
        raise NotImplementedError(("Writing out JMA files is not implemented ",
                                   "yet but is planned for a future release."))
        try:
            with open(path, 'w') as data_file:
                for n in range(self.t.shape[0]):
                    data_file.write("".join(("%s" % self.seconds2date(self.t[n]),
                                             " " * 4,
                                             "%s" % (int(self.eye_location[n, 0] *
                                                         10.0)),
                                             ", ",
                                             "%s" % (int(self.eye_location[n, 1] *
                                                         10.0)),
                                             ", ",
                                             "%s" % self.max_wind_speed[n],
                                             ", ",
                                             "%s" % self.central_pressure[n],
                                             ", ",
                                             ", " * 8,
                                             "%s" % self.storm_radius[n],
                                             ", ",
                                             "%s" % self.max_wind_radius[n],
                                             ", " * 10,
                                             "\n")))
        except Exception as e:
            # Remove possiblly partially generated file if not successful
            if os.path.exists(path):
                os.remove(path)
            raise e

    def write_imd(self, path, verbose=False):
        r"""Write out an IMD formatted storm file

        :Input:
         - *path* (string) Path to the file to be written.
         - *verbose* (bool) Print out additional information when writing.
        """
        raise NotImplementedError(("Writing out IMD files is not implemented ",
                                   "yet but is planned for a future release."))

    def write_tcvitals(self, path, verbose=False):
        r"""Write out an TCVITALS formatted storm file

        :Input:
         - *path* (string) Path to the file to be written.
         - *verbose* (bool) Print out additional information when writing.
         """

        raise NotImplementedError(("Writing in TCVITALS files is not",
                                   "implemented yet but is planned for a ",
                                   "future release."))

    # ================
    #  Track Plotting
    # ================
    def plot(self, ax, radius=None, t_range=None, coordinate_system=2, track_style='ko--',
             categorization="NHC", fill_alpha=0.25, fill_color='red'):
        """TO DO:  Write doc-string"""

        import matplotlib.pyplot as plt

        if isinstance(track_style, str):
            style = track_style

        # Extract information for plotting the track/swath
        t = self.t
        x = self.eye_location[:, 0]
        y = self.eye_location[:, 1]
        if t_range is not None:
            t = np.ma.masked_outside(t, t_range[0], t_range[1])
            x = np.ma.array(x, mask=t.mask).compressed()
            y = np.ma.array(y, mask=t.mask).compressed()
            t = t.compressed()

        # Plot track
        if isinstance(track_style, str):
            # Plot the track as a simple line with the given style
            ax.plot(x, y, track_style)
        elif isinstance(track_style, dict):
            if self.max_wind_speed is None:
                raise ValueError("Maximum wind speed not available so "
                                 "plotting catgories is not available.")

            # Plot the track using the colors provided in the dictionary
            cat_color_defaults = {5: 'red', 4: 'yellow', 3: 'orange', 2: 'green',
                                  1: 'blue', 0: 'gray', -1: 'lightgray'}
            colors = [track_style.get(category, cat_color_defaults[category])
                      for category in self.category(categorization=categorization)]
            for i in range(t.shape[0] - 1):
                ax.plot(x[i:i+2], y[i:i+2], color=colors[i], marker="o")

        else:
            raise ValueError("The `track_style` should be a string or dict.")

        # Plot swath
        if (isinstance(radius, float) or isinstance(radius, np.ndarray)
                or radius is None):

            if radius is None:
                # Default behavior
                if self.storm_radius is None:
                    raise ValueError("Cannot use storm radius for plotting "
                                     "the swath as the data is not available.")
                else:
                    if coordinate_system == 1:
                        _radius = self.storm_radius
                    elif coordinate_system == 2:
                        _radius = units.convert(self.storm_radius,
                                                'm', 'lat-long')
                    else:
                        raise ValueError(f"Unknown coordinate system "
                                         f"{coordinate_system} provided.")

            elif isinstance(radius, float):
                # Only one value for the radius was given, replicate
                _radius = np.ones(self.t.shape) * radius
            elif isinstance(radius, np.ndarray):
                # The array passed is the array to use
                _radius = radius
            else:
                raise ValueError("Invalid input argument for radius.  Should "
                                 "be a float or None")

            # Draw first and last points
            ax.add_patch(plt.Circle(
                (x[0], y[0]), _radius[0], color=fill_color))
            if t.shape[0] > 1:
                ax.add_patch(plt.Circle((x[-1], y[-1]), _radius[-1],
                                        color=fill_color))

            # Draw path around inner points
            if t.shape[0] > 2:
                for i in range(t.shape[0] - 1):
                    p = np.array([(x[i], y[i]), (x[i + 1], y[i + 1])])
                    v = p[1] - p[0]
                    if abs(v[1]) > 1e-16:
                        n = np.array([1, -v[0] / v[1]], dtype=float)
                    elif abs(v[0]) > 1e-16:
                        n = np.array([-v[1] / v[0], 1], dtype=float)
                    else:
                        raise Exception("Zero-vector given")
                    n /= np.linalg.norm(n)
                    n *= _radius[i]

                    ax.fill((p[0, 0] + n[0], p[0, 0] - n[0],
                             p[1, 0] - n[0],
                             p[1, 0] + n[0]),
                            (p[0, 1] + n[1], p[0, 1] - n[1],
                             p[1, 1] - n[1],
                             p[1, 1] + n[1]),
                            facecolor=fill_color, alpha=fill_alpha)
                    ax.add_patch(plt.Circle((p[1][0], p[1, 1]), _radius[i],
                                            color=fill_color, alpha=fill_alpha))

    # =========================================================================
    # Other Useful Routines

    def category(self, categorization="NHC", cat_names=False):
        r"""Categorizes storm based on relevant storm data

        :Input:
         - *categorization* (string) Type of categorization to use.  Defaults
           to the National Hurricane Center "NHC".
         - *cat_names* (bool) If True returns the category name rather than a
           number.  Default to *False*.

        :Output:
         - (ndarray) Integer array of categories at each time point of the
           storm.
         - (list) Similar to the above but the name of the category as a
           *string*.  This is only returned if *car_names = True*.

        """

        # TODO:  Need to standardize on 1-minute (almost never available) or
        # 10-minute (widely available) - see
        # https://en.wikipedia.org/wiki/Tropical_cyclone#Major_basins_and_related_warning_centers

        if categorization.upper() == "BEAUFORT":
            # Beaufort scale below uses knots
            speeds = units.convert(self.max_wind_speed, "m/s", "knots")
            category = (np.zeros(speeds.shape) +
                        (speeds >= 1) * (speeds < 4) * 1 +
                        (speeds >= 4) * (speeds < 7) * 2 +
                        (speeds >= 7) * (speeds < 11) * 3 +
                        (speeds >= 11) * (speeds < 17) * 4 +
                        (speeds >= 17) * (speeds < 22) * 5 +
                        (speeds >= 22) * (speeds < 28) * 6 +
                        (speeds >= 28) * (speeds < 34) * 7 +
                        (speeds >= 34) * (speeds < 41) * 8 +
                        (speeds >= 41) * (speeds < 48) * 9 +
                        (speeds >= 48) * (speeds < 56) * 10 +
                        (speeds >= 56) * (speeds < 64) * 11 +
                        (speeds >= 64) * 12)
            cat_map = {0: "Calm",
                       1: "Light air",
                       2: "Light breeze",
                       3: "Gentle breeze",
                       4: "Moderate breeze",
                       5: "Fresh breeze",
                       6: "Strong breeze",
                       7: "High wind",
                       8: "Gale",
                       9: "Strong gale",
                       10: "Whole gale",
                       11: "Violent storm",
                       12: "Hurricane"}

        elif categorization.upper() == "NHC":
            # TODO:  Change these to m/s (knots are how these are defined).
            # Definitely not in the correct format now
            # TODO:  Add TD and TS designations
            speeds = units.convert(self.max_wind_speed, "m/s", "knots")
            category = (np.zeros(speeds.shape) +
                        (speeds < 30) * -1 +
                        (speeds >= 64) * (speeds < 83) * 1 +
                        (speeds >= 83) * (speeds < 96) * 2 +
                        (speeds >= 96) * (speeds < 113) * 3 +
                        (speeds >= 113) * (speeds < 135) * 4 +
                        (speeds >= 135) * 5)
            cat_map = {-1: "Tropical Depression",
                       0: "Tropical Storm",
                       1: "Category 1 Hurricane",
                       2: "Category 2 Hurricane",
                       3: "Category 3 Hurricane",
                       4: "Category 4 Hurricane",
                       5: "Category 5 Hurricane"}

        elif categorization.upper() == "JTWC":
            raise NotImplementedError("JTWC categorization not implemented.")
        elif categorization.upper() == "JMA":
            raise NotImplementedError("JMA categorization not implemented.")
        elif categorization.upper() == "IMD":
            raise NotImplementedError("IMD categorization not implemented.")
        elif categorization.upper() == "MF":
            raise NotImplementedError("MF categorization not implemented.")
        elif categorization.upper() == "BOM":
            raise NotImplementedError("BOM categorization not implemented.")
        else:
            raise ValueError("Categorization %s not available."
                             % categorization)

        if cat_names:
            category_name = []
            for (i, cat) in enumerate(category):
                category_name.append(cat_map[cat])

            return category, category_name
        else:
            return category


# =============================================================================
# Model field construction - Models supported are
#  - Holland 1980 ('HOLLAND_1980') [1]
#  - Holland 2010 ('HOLLAND_2010') [2]
#  - Chavas, Lin, Emmanuel ('CLE_2015') [3]
# *TODO* - Add citations

# Dictionary of models.  Keys are function names, values are the proper name
# and a citation to the model
_supported_models = {"holland_1980": ["Holland 1980", "Holland, G. J. An Analytic Model of the Wind and Pressure Profiles in Hurricanes. Monthly Weather Review 108, 1212-1218 (1980)."],
                     "holland_2010": ["Holland 2010", "Holland, G. J., Belanger, J. I. & Fritz, A. A Revised Model for Radial Profiles of Hurricane Winds. Monthly Weather Review 138, 4393-4393 (2010)."],
                     "cle_2015": ["Chavas, Lin, Emmanuel 2015", "Chavas, D. R., Lin, N. & Emanuel, K. A Model for the Complete Radial Structure of the Tropical Cyclone Wind Field. Part I: Comparison with Observed Structure*. https://doi.org.ezproxy.cul.columbia.edu/10.1175/JAS-D-15-0014.1 72, 3647-3662 (2015)."]}


# In the case where the field is not rotationally symmetric then the r value
# defines the x and y axis extents.
def construct_fields(storm, r, t, model="holland_1980"):
    r""""""

    if model.lower() not in _supported_models.keys():
        raise ValueError("Model %s not available." % model)

    return getattr(sys.modules[__name__], model.lower())(storm, x, t)


# Specific implementations
def holland_1980(storm, r, t):
    r""""""
    raise NotImplementedError("Holland 1980 model has not been implemeted.")
    return None, None


def holland_2010(storm, r, t):
    r""""""
    raise NotImplementedError("Holland 2010 model has not been implemeted.")
    return None, None


def cle_2015(storm, r, t):
    r""""""
    raise NotImplementedError("CLE 2015 model has not been implemeted.")
    return None, None


# =============================================================================
# Radius fill functions
def fill_rad_w_other_source(t, storm_targ, storm_fill, var, interp_kwargs={}):
    r"""Fill in storm radius variable (*max_wind_radius* or \
    *storm_radius*) with values from another source. i.e.
    if you have missing radii in IBTrACS, you can fill with ATCF.
    This function will assume *storm_fill* has more non-missing
    values than *storm_targ* for this particular radius variable.
    Thus, it first attempts to interpolate the variable in *storm_fill*
    to the desired timestep. If that is missing, it tries to interpolate
    the non-missing values of the variable in *storm_targ*. If that
    also fails, it simply returns -1. The proper usage of this
    function is to wrap it such that you can pass a function
    with (*t*, *storm*) arguments to *max_wind_radius_fill* or
    *storm_radius_fill* when calling *write_geoclaw*.

    :Input:
    - *t* (:py:class:`datetime.datetime`) the time corresponding to
        a missing value of *max_wind_radius* or *storm_radius*
    - *storm_targ* (:py:class:`clawpack.geoclaw.storm.Storm`) storm
        that has missing values you want to fill
    - *storm_fill* (:py:class:`clawpack.geoclaw.storm.Storm`) storm
        that has non-missing values you want to use to fill *storm_targ*
    - *var* (str) Either 'max_wind_radius' or 'storm_radius'
    - *interp_kwargs* (dict) Additional keywords passed to scipy's
        interpolator.

    :Returns:
    - (float) value to use to fill this time point in *storm_targ*. -1
        if still missing after using *storm_fill* to fill.

    :Examples:

    .. code-block:: python

        >>> storm_ibtracs = Storm(file_format='IBTrACS', path='path_to_ibtracs.nc',
        ...     sid='2018300N26315')

        >>> storm_atcf = Storm(file_format='ATCF', path='path_to_atcf.dat')

        >>> def fill_mwr(t, storm):
        ...     return fill_rad_w_other_source(t, storm, storm_atcf, 'max_wind_radius')

        >>> storm_ibtracs.write(file_format = 'geoclaw',
        ...     path = 'out_path.storm',
        ...     max_wind_radius_fill = fill_mwr)
    """

    try:
        import xarray as xr
    except ImportError as e:
        print("fill_rad_w_other_source currently requires xarray to work.")
        raise e

    fill_da = xr.DataArray(getattr(storm_fill, var),
                           coords={'t': getattr(storm_fill, 't')},
                           dims=('t',))

    # convert -1 to nan
    fill_da = fill_da.where(fill_da > 0, np.nan)

    # if not all missing, try using storm_fill to fill
    if fill_da.notnull().any():

        # remove duplicates
        fill_da = fill_da.groupby('t').first()

        # remove NaNs
        fill_da = fill_da.dropna('t')

        # interpolate to point
        fill_interp = fill_da.interp(t=[t], kwargs=interp_kwargs).item()

        # try replacing with storm_fill
        # (assuming atcf has more data points than ibtracs)
        if not np.isnan(fill_interp):
            return fill_interp

    # next, try just interpolating other ibtracs values
    targ_da = xr.DataArray(getattr(storm_targ, var),
                           coords={'t': getattr(storm_targ, 't')},
                           dims=('t',))
    targ_da = targ_da.where(targ_da > 0, np.nan)
    if targ_da.notnull().any():
        targ_da = targ_da.groupby('t').first()
        targ_da = targ_da.dropna('t')
        targ_interp = targ_da.interp(t=[t], kwargs=interp_kwargs).item()
        if not np.isnan(targ_interp):
            return targ_interp

    # if nothing worked, return the missing value (-1)
    return -1


# =============================================================================
# Utility functions
def available_formats():
    r"""Construct a string suitable for listing available storm file formats.
    """
    output = "Available Formats: (Function, Name, Citation)\n"
    for (model, values) in Storm._supported_formats.items():
        output = "".join((output, "%s: %s %s\n" % (values[0], model,
                                                   values[1])))
    return output


def available_models():
    r"""Construct a string suitable for listing available storm models.
    """
    output = "Function, Name, Citation\n"
    for (model, values) in _supported_models.items():
        output = "".join((output, "%s: %s %s\n" % (values[0], model,
                                                   values[1])))
    return output


def make_multi_structure(path):
    r"""Create a dictionary of Storm objects for ATCF files with multiple storm tracks in them
    """
    with open(path, 'r') as f:
        lines = f.readlines()
        curTime = "test"
        curTrack = "test"
        os.mkdir("Clipped_ATCFs")
        stormDict = {}
        for line in lines:
            lineArr = line.split(", ")
            if curTime in lineArr[2]:
                if curTrack in lineArr[4]:
                    fileWrite.writelines(line)
                else:
                    fileWrite.close()
                    stormDict[curTime].update({curTrack: Storm(path=os.path.join(os.path.expandvars(
                        os.getcwd()), "Clipped_ATCFs", curTime, curTrack), file_format="ATCF")})
                    curTrack = lineArr[4]
                    fileWrite = open("Clipped_ATCFs/" +
                                     curTime + "/" + curTrack, 'w')
                    fileWrite.writelines(line)
            else:
                if curTime != "test":
                    fileWrite.close()
                    stormDict[curTime].update({curTrack: Storm(path=os.path.join(os.path.expandvars(
                        os.getcwd()), "Clipped_ATCFs", curTime, curTrack), file_format="ATCF")})
                curTime = lineArr[2]
                curTrack = lineArr[4]
                stormDict[curTime] = {}
                os.mkdir("Clipped_ATCFs/" + curTime)
                fileWrite = open("Clipped_ATCFs/" +
                                 curTime + "/" + curTrack, 'w')
                fileWrite.writelines(line)
    return stormDict


if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    # Positional argument
    parser.add_argument("path", help="Path to storm file to be read in")

    # Optional arguments
    parser.add_argument("-f", "--from", default="atcf", dest="input_format",
                        help="Format to convert from, defaults to 'atcf'")
    parser.add_argument("-o", "--output", default="geoclaw.storm",
                        dest="output_path",
                        help="Output path, default to 'geoclaw.storm'")
    parser.add_argument("-t", "--to", default="geoclaw",
                        dest="output_format",
                        help="Format to convert to, defaults to 'geoclaw'")
    parser.add_argument("-v", "--verbose",
                        help="Increase verbosity of output",
                        action="store_true")

    args = parser.parse_args()
    input_storm = Storm(args.path, file_format=args.input_format,
                        verbose=args.verbose)
    input_storm.write(args.output_path, file_format=args.output_format,
                      verbose=args.verbose)



import pytest

@pytest.mark.skip(reason="Test not working yet PeanoClaw.")
def test_initialization():
    from pyclaw.clawpack.clawpack import ClawSolver2D
    
    solver = ClawSolver2D()
    
    import peanoclaw
    peano_solver = peanoclaw.Solver(solver, 1.0)
    
    import inspect
    for member in inspect.getmembers(peano_solver):
        if(not member[0].startswith("_") and not inspect.ismethod(member[1])):
            print((member[0]))
    for member in inspect.getmembers(solver):
        if(not member[0].startswith("_") and not inspect.ismethod(member[1])):
            print((member[0]))
    

# if __name__=="__main__":
#     import nose
#     nose.main()



def test_acoustics_2d_variable_io():
    """Test I/O on variable-coefficient 2D acoustics application"""
    import os
    from . import acoustics_2d_interface
    thisdir = os.path.dirname(__file__)

    def verify_acoustics_io(controller):
        """ Verifies I/O on 2d variable-coefficient acoustics application"""
        from clawpack.pyclaw.util import check_diff
        from clawpack.pyclaw import Solution
        
        verify_dir = os.path.join(thisdir,'./io_test_verification')
        
        # Expected solution
        sol_0_expected = Solution()
        sol_0_expected.read(0,path=verify_dir,file_format='ascii',
                               file_prefix=None,read_aux=True)
        expected_aux = sol_0_expected.state.aux

        sol_20_expected = Solution()
        sol_20_expected.read(20,path=verify_dir,file_format='ascii',
                               file_prefix=None,read_aux=False)
        expected_q = sol_20_expected.state.q

        # Test solution
        sol_0_test = Solution()
        sol_0_test.read(0,path=controller.outdir,
                        file_format=controller.output_format,
                        file_prefix=None,read_aux=True,
                        options=controller.output_options)
        test_aux = sol_0_test.state.get_aux_global()

        sol_20_test = Solution()
        sol_20_test.read(20,path=controller.outdir,
                        file_format=controller.output_format,
                        file_prefix=None,read_aux=False,
                        options=controller.output_options)
        test_q = sol_20_test.state.get_q_global()


        if test_q is not None:
            q_err = check_diff(expected_q, test_q, reltol=1e-6)
            if q_err is not None:
                return q_err
        else:
            return

        if test_aux is not None:
            aux_err = check_diff(expected_aux, test_aux, reltol=1e-6)
            if aux_err is not None:
                return aux_err
        else:
            return
    
    tempdir = os.path.join(thisdir,'_io_test_results')
    claw = acoustics_2d_interface.setup(outdir=tempdir,num_cells=(50,50))
    claw.run()
    verify_acoustics_io(claw)



"""
Fix Makefile for changes to gauges_module.f90: remove dumpgauges.f
on amrclaw@77612a61b3
"""

from __future__ import absolute_import
from __future__ import print_function
import os

f = open('Makefile').read()
f = f.replace('  $(AMRLIB)/dumpgauge.f \\\n','')

os.system('mv Makefile Makefile_orig')
open('Makefile','w').write(f)
print('Modified Makefile, original in Makefile_orig')




r"""
Module for the CFL object.
"""
from __future__ import absolute_import
class CFL(object):
    """ Parallel CFL object, responsible for computing the
    Courant-Friedrichs-Lewy condition across all processes.
    """

    def __init__(self, global_max):
        from petsc4py import PETSc
        self._local_max = global_max
        self._global_max = global_max
        self._reduce_vec = PETSc.Vec().createWithArray([0])
        
    def get_global_max(self):
        r"""
        Compute the maximum CFL number over all processes for the current step.

        This is used to determine whether the CFL condition was
        violated and adjust the timestep.
        """
        self._reduce_vec.array = self._local_max
        self._global_max = self._reduce_vec.max()[1]
        return self._global_max

    def get_cached_max(self):
        return self._global_max

    def set_local_max(self,new_local_max):
        self._local_max = new_local_max

    def set_global_max(self,new_global_max):
        self._global_max = new_global_max

    def update_global_max(self,new_local_max):
        self._reduce_vec.array = new_local_max
        self._global_max = max(self._global_max,self._reduce_vec.max()[1])

